# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import logging
import time
from pathlib import Path

from gi.repository import GdkPixbuf
from gi.repository import GLib
from gi.repository import Gtk

from gajim.common import app
from gajim.common import ged
from gajim.common.const import AvatarSize
from gajim.common.events import FileCompleted
from gajim.common.events import FileError
from gajim.common.events import FileHashError
from gajim.common.events import FileProgress
from gajim.common.events import FileRequestError
from gajim.common.events import FileRequestReceivedEvent
from gajim.common.events import FileRequestSent
from gajim.common.events import FileSendError
from gajim.common.events import JingleErrorReceived
from gajim.common.events import JingleFtCancelledReceived
from gajim.common.file_props import FileProp
from gajim.common.file_props import FilesProp
from gajim.common.ged import EventHelper
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.storage.archive import models as mod
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.util.datetime import utc_now
from gajim.common.util.text import format_eta

from gajim.gtk.builder import get_builder
from gajim.gtk.conversation.rows.base import BaseRow
from gajim.gtk.conversation.rows.widgets import DateTimeLabel
from gajim.gtk.conversation.rows.widgets import NicknameLabel
from gajim.gtk.util.misc import open_file

TransferEventT = FileRequestReceivedEvent | FileRequestSent

log = logging.getLogger("gajim.gtk.conversation.rows.file_transfer_jingle")


class FileTransferJingleRow(BaseRow, EventHelper):
    def __init__(
        self,
        account: str,
        contact: BareContact,
        event: TransferEventT | None = None,
        message: mod.Message | None = None,
    ) -> None:
        BaseRow.__init__(self, account)
        EventHelper.__init__(self)

        self.type = "file-transfer"

        if message is not None:
            timestamp = message.timestamp
        else:
            timestamp = utc_now()
        self.timestamp = timestamp.astimezone()
        self.db_timestamp = timestamp.timestamp()

        self._contact = contact
        self._file_props = None

        if message is not None and message.filetransfers:
            # TODO: Handle filetransfers and sources specifically
            file_transfer = message.filetransfers[0]
            sources = file_transfer.source
            for source in sources:
                if isinstance(source, mod.JingleFT):
                    self._file_props = FilesProp.getFilePropBySid(source.sid)
                    if self._file_props is None:
                        log.debug("File prop not found for SID: %s", source.sid)
            self.pk = message.pk
        else:
            assert event is not None
            self._file_props = event.file_props
        self._start_time = 0

        if app.settings.get("use_kib_mib"):
            self._units = GLib.FormatSizeFlags.IEC_UNITS
        else:
            self._units = GLib.FormatSizeFlags.DEFAULT

        avatar_placeholder = Gtk.Box()
        avatar_placeholder.set_size_request(AvatarSize.ROSTER, -1)
        avatar_placeholder.set_valign(Gtk.Align.START)
        self.grid.attach(avatar_placeholder, 0, 0, 1, 1)

        if message is not None:
            if message.direction == ChatDirection.INCOMING:
                contact = self._contact
                is_self = True
            else:
                bare_contact = self._client.get_module("Contacts").get_contact(
                    self._client.get_own_jid().bare
                )
                assert isinstance(bare_contact, BareContact)
                contact = bare_contact
                is_self = False
        else:
            if isinstance(event, FileRequestSent):
                bare_contact = self._client.get_module("Contacts").get_contact(
                    self._client.get_own_jid().bare
                )
                assert isinstance(bare_contact, BareContact)
                contact = bare_contact
                is_self = False
            else:
                contact = self._contact
                is_self = True

        scale = self.get_scale_factor()
        avatar = contact.get_avatar(AvatarSize.ROSTER, scale, add_show=False)
        assert not isinstance(avatar, GdkPixbuf.Pixbuf)
        avatar_image = Gtk.Image.new_from_paintable(avatar)
        avatar_placeholder.append(avatar_image)

        name_widget = NicknameLabel(contact.name, is_self)
        name_widget.set_halign(Gtk.Align.START)
        name_widget.set_valign(Gtk.Align.START)

        timestamp_widget = DateTimeLabel(self.timestamp)
        timestamp_widget.set_hexpand(True)
        timestamp_widget.set_valign(Gtk.Align.START)

        meta_box = Gtk.Box()
        meta_box.set_spacing(6)
        meta_box.append(name_widget)
        meta_box.append(timestamp_widget)
        self.grid.attach(meta_box, 1, 0, 1, 1)

        self._ui = get_builder("file_transfer_jingle.ui")
        self.grid.attach(self._ui.transfer_box, 1, 1, 1, 1)
        self._ui.transfer_box.set_halign(Gtk.Align.START)

        self._connect(
            self._ui.accept_file_request, "clicked", self._on_accept_file_request
        )
        self._connect(
            self._ui.reject_file_request, "clicked", self._on_reject_file_request
        )
        self._connect(self._ui.open_folder, "clicked", self._on_open_folder)
        self._connect(self._ui.open_file, "clicked", self._on_open_file)
        self._connect(self._ui.error_show_transfers, "clicked", self._on_show_transfers)
        self._connect(self._ui.retry_bad_hash, "clicked", self._on_bad_hash_retry)
        self._connect(
            self._ui.rejected_show_transfers, "clicked", self._on_show_transfers
        )
        self._connect(self._ui.cancel_transfer, "clicked", self._on_cancel_transfer)

        if message is not None:
            self._reconstruct_transfer()
        else:
            assert event is not None
            self._display_transfer_info(event.name)

        if self._file_props is None:
            return

        self.register_events(
            [
                ("file-completed", ged.GUI1, self.process_event),
                ("file-hash-error", ged.GUI1, self.process_event),
                ("file-send-error", ged.GUI1, self.process_event),
                ("file-request-error", ged.GUI1, self.process_event),
                ("file-progress", ged.GUI1, self.process_event),
                ("file-error", ged.GUI1, self.process_event),
                ("jingle-error-received", ged.GUI1, self.process_event),
                ("jingle-ft-cancelled-received", ged.GUI1, self.process_event),
            ]
        )

    def do_unroot(self) -> None:
        self.unregister_events()
        BaseRow.do_unroot(self)

    def _reconstruct_transfer(self) -> None:
        self._show_file_infos()
        if self._file_props is None:
            self._ui.transfer_action.set_text(_("File Transfer"))
            self._ui.error_label.set_text(_("No info available"))
            self._ui.action_stack.set_visible_child_name("error")
            return

        if self._file_props.completed:
            self._show_completed()
            return

        if self._file_props.stopped:
            self._ui.action_stack.set_visible_child_name("error")
            self._ui.transfer_action.set_text(_("File Transfer Stopped"))
            self._ui.error_label.set_text("")
            return

        if self._file_props.error is not None:
            self._show_error(self._file_props)
            return

        self._ui.transfer_action.set_text(_("File Offered…"))

    def _display_transfer_info(self, event_name: str) -> None:
        if event_name == "file-request-sent":
            self._ui.action_stack.set_visible_child_name("progress")
            self._ui.progress_label.set_text(_("Waiting…"))

        self._ui.transfer_action.set_text(_("File Offered…"))
        self._show_file_infos()

    def _show_file_infos(self) -> None:
        if self._file_props is None:
            self._ui.file_name.set_visible(False)
            self._ui.file_description.set_visible(False)
            self._ui.file_size.set_visible(False)
            return

        file_name = GLib.markup_escape_text(str(self._file_props.name))
        if self._file_props.mime_type:
            file_name = f"{file_name} ({self._file_props.mime_type})"
        self._ui.file_name.set_text(file_name)
        self._ui.file_name.set_tooltip_text(file_name)

        if self._file_props.desc:
            desc = GLib.markup_escape_text(self._file_props.desc)
            self._ui.file_description.set_text(desc)
            self._ui.file_description.set_tooltip_text(desc)
        else:
            self._ui.file_description.set_visible(False)

        assert self._file_props.size is not None
        self._ui.file_size.set_text(
            GLib.format_size_full(self._file_props.size, self._units)
        )

    def process_event(self, event: TransferEventT) -> None:
        assert self._file_props is not None

        if isinstance(event, JingleErrorReceived):
            if event.sid != self._file_props.sid:
                return
            self._ui.action_stack.set_visible_child_name("error")
            self._ui.transfer_action.set_text(_("File Transfer Cancelled"))
            self._ui.error_label.set_text(event.reason)
            return

        if isinstance(event, JingleFtCancelledReceived):
            if event.sid != self._file_props.sid:
                return
            self._ui.action_stack.set_visible_child_name("error")
            self._ui.transfer_action.set_text(_("File Transfer Cancelled"))
            self._ui.error_label.set_text(
                _("%(name)s cancelled the transfer (%(reason)s)")
                % {"name": self._contact.name, "reason": event.reason}
            )
            return

        if event.file_props.sid != self._file_props.sid:
            return

        if isinstance(event, FileCompleted):
            self._show_completed()
        elif isinstance(event, FileError):
            self._show_error(event.file_props)
        elif isinstance(event, FileHashError):
            self._ui.action_stack.set_visible_child_name("hash-error")
            self._ui.transfer_action.set_text(_("File Verification Failed"))
        elif isinstance(event, FileRequestError | FileSendError):
            self._ui.action_stack.set_visible_child_name("error")
            self._ui.transfer_action.set_text(_("File Transfer Cancelled"))
            error_text = (
                _("Connection with %s could not be established.") % self._contact.name
            )
            if event.error_msg:
                error_text = f"{error_text} ({event.error_msg})"
            self._ui.error_label.set_text(error_text)

        elif isinstance(event, FileProgress):
            self._update_progress(event.file_props)

    def _update_progress(self, file_props: FileProp) -> None:
        self._ui.action_stack.set_visible_child_name("progress")
        self._ui.transfer_action.set_text(_("Transferring File…"))

        time_now = time.time()
        full_size = file_props.size
        assert full_size is not None

        if file_props.type_ == "s":
            # We're sending a file
            if self._start_time == 0:
                self._start_time = time_now
                return
            if not file_props.transferred_size:
                return
            transferred_size = file_props.transferred_size[-1][1]
        else:
            # We're receiving a file
            transferred_size = file_props.received_len
            assert transferred_size is not None

        if full_size == 0:
            return

        bytes_sec = int(round(transferred_size / (time_now - self._start_time), 1))
        speed = f"{GLib.format_size_full(bytes_sec, self._units)}/s"
        self._ui.progress_label.set_tooltip_text(_("Speed: %s") % speed)

        if bytes_sec == 0:
            eta = "∞"
        else:
            eta = format_eta(round((full_size - transferred_size) / bytes_sec))

        progress = float(transferred_size) / full_size
        self._ui.progress_label.set_text(
            _("%(progress)s %% (%(time)s remaining)")
            % {"progress": round(progress * 100), "time": eta}
        )

        self._ui.progress_bar.set_fraction(progress)

    def _show_error(self, file_props: FileProp) -> None:
        self._ui.action_stack.set_visible_child_name("error")
        self._ui.transfer_action.set_text(_("File Transfer Stopped"))
        if file_props.error == -1:
            self._ui.error_label.set_text(
                _("%s stopped the transfer") % self._contact.name
            )
        elif file_props.error == -6:
            self._ui.error_label.set_text(_("Error opening file"))
        elif file_props.error == -12:
            self._ui.error_label.set_text(_("SSL certificate error"))
        else:
            self._ui.error_label.set_text(_("An error occurred"))

    def _show_completed(self) -> None:
        self._ui.action_stack.set_visible_child_name("complete")
        self._ui.transfer_action.set_text(_("File Transfer Completed"))

    def _on_accept_file_request(self, _button: Gtk.Button) -> None:
        pass
        # app.interface.instances["file_transfers"].on_file_request_accepted(
        #     self._account, self._contact, self._file_props
        # )
        # self._start_time = time.time()

    def _on_reject_file_request(self, _button: Gtk.Button) -> None:
        assert self._file_props is not None
        self._client.get_module("Bytestream").send_file_rejection(self._file_props)
        assert self._file_props is not None
        self._file_props.stopped = True
        self._ui.action_stack.set_visible_child_name("rejected")
        self._ui.transfer_action.set_text(_("File Transfer Cancelled"))

    def _on_open_file(self, _button: Gtk.Button) -> None:
        assert self._file_props is not None
        assert self._file_props.file_name is not None
        open_file(Path(self._file_props.file_name))

    def _on_open_folder(self, _button: Gtk.Button) -> None:
        assert self._file_props is not None
        assert self._file_props.file_name is not None
        open_file(Path(self._file_props.file_name), show_in_folder=True)

    def _on_bad_hash_retry(self, _button: Gtk.Button) -> None:
        pass
        # app.interface.instances["file_transfers"].show_hash_error(
        #     self._contact.jid, self._file_props, self._account
        # )

    def _on_cancel_transfer(self, _button: Gtk.Button) -> None:
        pass
        # app.interface.instances["file_transfers"].cancel_transfer(self._file_props)

    def _on_show_transfers(self, _button: Gtk.Button) -> None:
        pass
        # file_transfers = app.interface.instances["file_transfers"]
        # if file_transfers.window.get_property("visible"):
        #     file_transfers.window.present()
        # else:
        #     file_transfers.window.show_all()



#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import cast

from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango

from gajim.common import app
from gajim.common.const import AvatarSize
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant

from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import iterate_children


class ContactItem(Gtk.FlowBoxChild, SignalManager):
    def __init__(self, account: str, jid: str, is_new: bool = False) -> None:
        Gtk.FlowBoxChild.__init__(self, width_request=150)
        SignalManager.__init__(self)

        self.account = account
        self.jid = jid
        self.is_new = is_new

        name_label = Gtk.Label()
        name_label.set_ellipsize(Pango.EllipsizeMode.END)
        name_label.set_max_width_chars(12)
        name_label.add_css_class("bold")

        if is_new:
            avatar_image = Gtk.Image.new_from_icon_name("lucide-user-symbolic")
            name_label.set_text(jid)
            name_label.set_tooltip_text(jid)
        else:
            client = app.get_client(account)
            contact = client.get_module("Contacts").get_contact(jid)
            assert isinstance(
                contact, BareContact | GroupchatContact | GroupchatParticipant
            )
            texture = contact.get_avatar(AvatarSize.ROSTER, self.get_scale_factor())
            avatar_image = Gtk.Image.new_from_paintable(texture)
            avatar_image.set_pixel_size(AvatarSize.ROSTER)
            name_label.set_text(contact.name)
            name_label.set_tooltip_text(contact.name)

        remove_button = Gtk.Button.new_from_icon_name("lucide-trash-symbolic")
        remove_button.set_valign(Gtk.Align.CENTER)
        remove_button.set_halign(Gtk.Align.END)
        remove_button.set_hexpand(True)
        remove_button.set_tooltip_text(_("Remove"))
        self._connect(remove_button, "clicked", self._on_remove)

        box = Gtk.Box(spacing=6)
        box.set_valign(Gtk.Align.CENTER)
        box.append(avatar_image)
        box.append(name_label)
        box.append(remove_button)
        box.add_css_class("contact-flowbox-item")
        self.set_child(box)

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.FlowBoxChild.do_unroot(self)
        app.check_finalize(self)

    def _on_remove(self, _button: Gtk.Button) -> None:
        flow_box = cast(ContactsFlowBox, self.get_parent())
        flow_box.on_contact_removed(self)


class ContactsFlowBox(Gtk.FlowBox):
    __gsignals__ = {
        "contact-removed": (GObject.SignalFlags.RUN_LAST, None, (str, str, bool))
    }

    def __init__(self) -> None:
        Gtk.FlowBox.__init__(self)
        self.set_column_spacing(6)
        self.set_row_spacing(3)
        self.set_selection_mode(Gtk.SelectionMode.NONE)
        self.set_valign(Gtk.Align.START)
        self.set_can_focus(False)

    def do_unroot(self) -> None:
        Gtk.FlowBox.do_unroot(self)
        app.check_finalize(self)

    def clear(self) -> None:
        self.remove_all()

    def add_contact(self, account: str, jid: str, is_new: bool = False) -> None:
        contact_item = ContactItem(account, jid, is_new=is_new)
        self.append(contact_item)

    def has_contacts(self) -> bool:
        return bool(self.get_child_at_index(0) is not None)

    def get_contact_jids(self) -> list[str]:
        contacts: list[str] = []
        contact_items = cast(list[ContactItem], iterate_children(self))
        for contact in contact_items:
            contacts.append(contact.jid)
        return contacts

    def on_contact_removed(self, row: ContactItem) -> None:
        account = row.account
        jid = row.jid
        is_new = row.is_new
        self.remove(row)
        self.emit("contact-removed", account, jid, is_new)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# Iq handler

from __future__ import annotations

import nbxmpp
from nbxmpp.structs import IqProperties
from nbxmpp.structs import StanzaHandler

from gajim.common import app
from gajim.common import types
from gajim.common.events import FileRequestError
from gajim.common.events import FileSendError
from gajim.common.file_props import FilesProp
from gajim.common.helpers import to_user_string
from gajim.common.modules.base import BaseModule


class Iq(BaseModule):
    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.handlers = [
            StanzaHandler(
                name="iq", callback=self._iq_error_received, typ="error", priority=51
            ),
        ]

    def _iq_error_received(
        self,
        _con: types.NBXMPPClient,
        _stanza: nbxmpp.protocol.Iq,
        properties: IqProperties,
    ) -> None:
        self._log.info("Error: %s", properties.error)
        assert properties.error is not None
        if properties.error.condition in (
            "jid-malformed",
            "forbidden",
            "not-acceptable",
        ):
            assert properties.id is not None
            sid = self._get_sid(properties.id)
            file_props = FilesProp.getFileProp(self._account, sid)
            if file_props:
                if properties.error.condition == "jid-malformed":
                    file_props.error = -3
                else:
                    file_props.error = -4

                assert properties.jid is not None
                app.ged.raise_event(
                    FileRequestError(
                        conn=self._con,
                        jid=properties.jid.bare,
                        file_props=file_props,
                        error_msg=to_user_string(properties.error),
                    )
                )
                self._con.get_module("Bytestream").disconnect_transfer(file_props)
                raise nbxmpp.NodeProcessed

        if properties.error.condition == "item-not-found":
            assert properties.id is not None
            sid = self._get_sid(properties.id)
            file_props = FilesProp.getFileProp(self._account, sid)
            if file_props:
                app.ged.raise_event(
                    FileSendError(
                        account=self._account,
                        jid=str(properties.jid),
                        file_props=file_props,
                        error_msg="",
                    )
                )
                self._con.get_module("Bytestream").disconnect_transfer(file_props)
                raise nbxmpp.NodeProcessed

        self._log.error("Received iq error with unknown id: %s", properties.error)

        raise nbxmpp.NodeProcessed

    @staticmethod
    def _get_sid(id_: str) -> str:
        sid = id_
        if len(id_) > 3 and id_[2] == "_":
            sid = id_[3:]
        return sid



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from enum import IntEnum


class AudioPlayerState(IntEnum):
    NULL = 0
    READY = 1
    PAUSED = 2
    PLAYING = 3


class FTState(IntEnum):
    CREATED = 0
    PREPARING = 1
    ENCRYPTING = 2
    DECRYPTING = 3
    STARTED = 4
    IN_PROGRESS = 5
    FINISHED = 6
    ERROR = 7
    CANCELLED = 8


class PreviewState(IntEnum):
    INIT = 0
    DOWNLOADING = 1
    OFFER_DOWNLOAD = 2
    DOWNLOADED = 3
    DISPLAY = 4
    ERROR = 5



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Any
from typing import TextIO

import cProfile
import json
import logging
import os
import pstats
import sys
from datetime import datetime
from pathlib import Path
from pstats import SortKey

from gi.repository import GLib
from nbxmpp.const import ConnectionProtocol
from nbxmpp.const import ConnectionType
from nbxmpp.protocol import JID
from packaging.version import Version as V

from gajim import IS_PORTABLE
from gajim.common import app
from gajim.common import configpaths
from gajim.common import ged
from gajim.common import logging_helpers
from gajim.common import passwords
from gajim.common.cert_store import CertificateStore
from gajim.common.client import Client
from gajim.common.commands import ChatCommands
from gajim.common.events import AccountCreated
from gajim.common.events import AccountDisabled
from gajim.common.events import AccountDisconnected
from gajim.common.events import AccountEnabled
from gajim.common.events import AccountRemoved
from gajim.common.events import AllowGajimUpdateCheck
from gajim.common.events import DBMigrationError
from gajim.common.events import GajimUpdateAvailable
from gajim.common.events import SignedIn
from gajim.common.file_transfer_manager import FileTransfer
from gajim.common.file_transfer_manager import FileTransferManager
from gajim.common.settings import Settings
from gajim.common.storage.archive.storage import MessageArchiveStorage
from gajim.common.storage.cache import CacheStorage
from gajim.common.storage.events.storage import EventStorage
from gajim.common.storage.openpgp.storage import OpenPGPStorage
from gajim.common.task_manager import PulseManager
from gajim.common.task_manager import TaskManager
from gajim.common.util.text import from_one_line
from gajim.common.util.text import get_random_string

log = logging.getLogger("gajim.c.application")


class CoreApplication(ged.EventHelper):
    def __init__(self) -> None:
        ged.EventHelper.__init__(self)
        self._profiling_session = None
        self._cmd_line_options: dict[str, Any] = {}

    def _init_core(self, *, in_memory: bool) -> bool:
        # Create and initialize Application Paths & Databases
        app.app = self
        app.print_version()
        configpaths.create_paths()

        configure_path = self._cmd_line_options.get("configure")

        app.settings = Settings(in_memory=in_memory)
        app.settings.init(configure_path)

        if configure_path:
            sys.exit()

        if export_path := self._cmd_line_options.get("export-settings"):
            app.settings.export_to_json(export_path)
            sys.exit()

        passwords.init()
        app.detect_dependencies()
        app.init_process_pool()

        app.commands = ChatCommands()

        try:
            app.storage.cache = CacheStorage(in_memory=in_memory)
            app.storage.cache.init()

            app.storage.events = EventStorage()
            app.storage.events.init()

            app.storage.archive = MessageArchiveStorage(in_memory=in_memory)
            app.storage.archive.init()

            app.storage.openpgp = OpenPGPStorage(in_memory=in_memory)
            app.storage.openpgp.init()
        except Exception as error:
            app.ged.raise_event(DBMigrationError(exception=error))
            log.exception("Failed to init storage")
            return False

        app.cert_store = CertificateStore()
        app.task_manager = TaskManager()
        app.pulse_manager = PulseManager()

        app.ftm = FileTransferManager()

        # from gajim.common.call_manager import CallManager
        # app.call_manager = CallManager()

        from gajim.gtk.audio_player import AudioPlayer

        app.audio_player = AudioPlayer()

        if sys.platform in ("win32", "darwin"):
            GLib.timeout_add_seconds(20, self._check_for_updates)

        for account in app.settings.get_active_accounts():
            app.connections[account] = Client(account)
            app.to_be_removed[account] = []

        from gajim.plugins.pluginmanager import PluginManager
        from gajim.plugins.repository import PluginRepository

        app.plugin_manager = PluginManager()
        app.plugin_manager.init_plugins()
        app.plugin_repository = PluginRepository()

        app.commands.init()

        for client in app.get_clients():
            client.get_module("Roster").load_roster()

        GLib.timeout_add_seconds(5, self._remote_init)

        self.register_events(
            [
                ("signed-in", ged.CORE, self._on_signed_in),
            ]
        )

        return True

    @property
    def _log(self) -> logging.Logger:
        return app.log("app")

    def _core_command_line(self, options: GLib.VariantDict) -> None:
        if options.contains("cprofile"):
            self.start_profiling()

        if options.contains("gdebug"):
            os.environ["G_MESSAGES_DEBUG"] = "all"

        config_path = options.lookup_value("config-path")
        if config_path is not None:
            config_path = config_path.get_string()
            configpaths.set_config_root(config_path)

        configpaths.init()
        logging_helpers.init()

        if options.contains("quiet"):
            logging_helpers.set_quiet()

        if options.contains("verbose"):
            logging_helpers.set_verbose()

        loglevel = options.lookup_value("loglevel")
        if loglevel is not None:
            loglevel = loglevel.get_string()
            logging_helpers.set_loglevels(loglevel)

        if options.contains("warnings"):
            self._show_warnings()

        configure_path = options.lookup_value("configure")
        if configure_path is not None:
            self._cmd_line_options["configure"] = Path(configure_path.get_string())

        export_path = options.lookup_value("export-settings")
        if export_path is not None:
            self._cmd_line_options["export-settings"] = Path(export_path.get_string())

    def _auto_connect(self) -> None:
        for client in app.get_clients():
            account = client.account
            if not app.settings.get_account_setting(account, "autoconnect"):
                continue

            status = "online"
            status_message = ""

            if app.settings.get_account_setting(account, "restore_last_status"):
                status = app.settings.get_account_setting(account, "last_status")
                status_message = app.settings.get_account_setting(
                    account, "last_status_msg"
                )
                status_message = from_one_line(status_message)

            client.change_status(status, status_message)

    def _remote_init(self) -> None:
        if not app.settings.get("remote_control"):
            return

        try:
            from gajim.common.dbus import remote_control

            remote_control.GajimRemote()
        except Exception:
            self._log.exception("Failed to init remote control")

    def start_profiling(self) -> None:
        self._log.info("Start profiling")
        self._profiling_session = cProfile.Profile()
        self._profiling_session.enable()

    def end_profiling(self) -> None:
        if self._profiling_session is None:
            return

        self._profiling_session.disable()
        self._log.info("End profiling")
        ps = pstats.Stats(self._profiling_session)
        ps = ps.sort_stats(SortKey.TIME)
        ps.print_stats()

    def _start_shutdown(self) -> None:
        self._log.info("Start Shutdown")
        accounts_to_disconnect: dict[str, Client] = {}

        for client in app.get_clients():
            if client.is_available():
                accounts_to_disconnect[client.account] = client

        if not accounts_to_disconnect:
            self._shutdown_core()
            self._shutdown_complete()
            return

        def _on_disconnect(event: AccountDisconnected) -> None:
            accounts_to_disconnect.pop(event.account)
            if not accounts_to_disconnect:
                self._shutdown_core()
                self._shutdown_complete()
                return

        app.ged.register_event_handler(
            "account-disconnected", ged.POSTGUI, _on_disconnect
        )

        for client in accounts_to_disconnect.values():
            client.change_status("offline", "")

    def _shutdown_core(self) -> None:
        self._log.info("Shutdown core")
        # Commit any outstanding SQL transactions
        app.process_pool.shutdown(cancel_futures=True)
        app.storage.archive.cleanup_chat_history()
        app.storage.cache.shutdown()
        app.storage.archive.shutdown()
        app.settings.save()
        app.settings.shutdown()

        self.end_profiling()
        configpaths.cleanup_temp()

    def _shutdown_complete(self) -> None:
        self._log.info("Shutdown complete")

    @staticmethod
    def _show_warnings() -> None:
        import traceback
        import warnings

        def warn_with_traceback(
            message: Warning | str,
            category: type[Warning],
            filename: str,
            lineno: int,
            _file: TextIO | None = None,
            line: str | None = None,
        ) -> None:

            traceback.print_stack(file=sys.stderr)
            sys.stderr.write(
                warnings.formatwarning(message, category, filename, lineno, line)
            )

        warnings.showwarning = warn_with_traceback
        warnings.filterwarnings(action="always")

    def _check_for_updates(self) -> None:
        if not app.settings.get("check_for_update"):
            return

        if sys.platform == "win32" and app.is_ms_store():
            # Gajim updates are handled by MS Store directly
            return

        now = datetime.now()
        last_check = app.settings.get("last_update_check")
        if not last_check:
            app.ged.raise_event(AllowGajimUpdateCheck())
            return

        last_check_time = datetime.strptime(last_check, "%Y-%m-%d %H:%M")
        if (now - last_check_time).days < 7:
            return

        self.check_for_gajim_updates()

    def check_for_gajim_updates(self) -> None:
        self._log.info("Checking for Gajim updates")

        app.ftm.http_request(
            "GET",
            "https://gajim.org/current-version.json",
            callback=self._on_update_response,
        )

    def _on_update_response(self, obj: FileTransfer) -> None:
        try:
            result = obj.get_result()
        except Exception as error:
            self._log.warning("Error while performing gajim update check: %s", error)
            return

        now = datetime.now()
        app.settings.set("last_update_check", now.strftime("%Y-%m-%d %H:%M"))

        try:
            data = json.loads(result.content)
            latest_version = data["current_version"]
            setup_url = data["current_win_setup"]
            if IS_PORTABLE:
                setup_url = data["current_win_portable_setup"]
        except Exception:
            self._log.exception("Unable to perform update check")
            return

        if V(latest_version) > V(app.version):
            app.ged.raise_event(
                GajimUpdateAvailable(version=latest_version, setup_url=setup_url)
            )
            return

        self._log.info("Gajim is up to date")

    def create_account(
        self,
        account: str,
        address: JID,
        password: str | None,
        proxy_name: str | None,
        custom_host: tuple[str, ConnectionProtocol, ConnectionType] | None,
        anonymous: bool = False,
    ) -> None:

        if anonymous:
            address = address.new_with(localpart="anon")

        if not address.localpart:
            raise ValueError("Username must be set")

        account_label = str(address)

        config: dict[str, str | int | bool] = {
            "address": str(address),
            "resource": f"gajim.{get_random_string(8)}",
            "account_label": account_label,
            "anonymous_auth": anonymous,
        }

        if proxy_name:
            config["proxy"] = proxy_name

        use_custom_host = custom_host is not None
        config["use_custom_host"] = use_custom_host
        if custom_host:
            host, _protocol, type_ = custom_host
            host, port = host.rsplit(":", maxsplit=1)
            config["custom_port"] = int(port)
            config["custom_host"] = host
            config["custom_type"] = type_.value

        app.settings.add_account(account)
        for opt, value in config.items():
            app.settings.set_account_setting(account, opt, value)  # type: ignore  # noqa: E501

        if not anonymous and password is not None:
            # Password module depends on existing config
            passwords.save_password(account, password)

        app.ged.raise_event(AccountCreated(account=account))

    def enable_account(self, account: str, *, connect: bool) -> None:
        app.connections[account] = Client(account)

        app.plugin_manager.register_modules_for_account(app.connections[account])

        app.to_be_removed[account] = []
        app.settings.set_account_setting(account, "active", True)

        app.ged.raise_event(AccountEnabled(account=account))

        if connect:
            app.get_client(account).change_status("online", "")

    def disable_account(self, account: str) -> None:
        app.settings.set_account_setting(account, "roster_version", "")
        app.settings.set_account_setting(account, "active", False)

        # Code in account-disabled handlers may use app.get_client()
        app.ged.raise_event(AccountDisabled(account=account))

        app.get_client(account).cleanup()
        del app.connections[account]
        del app.to_be_removed[account]

    def remove_account(self, account: str) -> None:
        if app.settings.get_account_setting(account, "active"):
            self.disable_account(account)

        app.storage.cache.remove_roster(account)
        # Delete password must be before del_per() because it calls set_per()
        # which would recreate the account with defaults values if not found
        passwords.delete_password(account)
        app.storage.archive.remove_account(account)
        app.storage.openpgp.remove_account(account)
        app.settings.remove_account(account)
        app.app.remove_account_actions(account)

        app.ged.raise_event(AccountRemoved(account=account))

    def _on_signed_in(self, event: SignedIn) -> None:
        client = app.get_client(event.account)
        if client.get_module("MAM").available:
            client.get_module("MAM").request_archive_on_signin()

    def change_status(self, status: str, account: str | None = None) -> None:

        if account is not None:
            self._change_status(account, status)
            return

        for client in app.get_clients():
            if not app.settings.get_account_setting(
                client.account, "sync_with_global_status"
            ):
                continue

            self._change_status(client.account, status)

    @staticmethod
    def _change_status(account: str, status: str) -> None:
        client = app.get_client(account)
        message = client.status_message

        if status == "offline":
            # TODO delete pep
            # self.delete_pep(app.get_jid_from_account(account), account)
            pass

        client.change_status(status, message)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import locale

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango

from gajim.common import app
from gajim.common import types
from gajim.common.const import AvatarSize
from gajim.common.const import Direction
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.util.jid import validate_jid

from gajim.gtk.builder import get_builder
from gajim.gtk.contacts_flowbox import ContactsFlowBox
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import iterate_listbox_children
from gajim.gtk.widgets import AccountBadge


class ContactRow(Gtk.ListBoxRow):
    def __init__(
        self,
        account: str,
        contact: types.BareContact | None,
        jid: str,
        name: str,
        show_account: bool,
    ) -> None:
        Gtk.ListBoxRow.__init__(self)
        self.add_css_class("start-chat-row")
        self.account = account
        self.jid = jid

        self._contact = contact
        self._name = name
        self._show_account = show_account

        self._account_label = app.get_account_label(account)
        self.is_new: bool = jid == ""

        grid = Gtk.Grid()
        grid.set_column_spacing(12)
        grid.set_size_request(260, -1)

        image = self._get_avatar_image(contact)
        image.set_size_request(AvatarSize.ROSTER, AvatarSize.ROSTER)
        grid.attach(image, 0, 0, 1, 1)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
        box.set_hexpand(True)

        self._name_label = Gtk.Label(label=self._name)
        self._name_label.set_ellipsize(Pango.EllipsizeMode.END)
        self._name_label.set_xalign(0)
        self._name_label.set_width_chars(20)
        self._name_label.set_halign(Gtk.Align.START)
        self._name_label.add_css_class("title-3")
        name_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        name_box.append(self._name_label)

        if show_account:
            account_badge = AccountBadge(account)
            account_badge.set_halign(Gtk.Align.END)
            account_badge.set_valign(Gtk.Align.START)
            account_badge.set_hexpand(True)
            name_box.append(account_badge)
        box.append(name_box)

        self._jid_label = Gtk.Label(label=str(jid))
        self._jid_label.set_tooltip_text(str(jid))
        self._jid_label.set_ellipsize(Pango.EllipsizeMode.END)
        self._jid_label.set_xalign(0)
        self._jid_label.set_width_chars(22)
        self._jid_label.set_halign(Gtk.Align.START)
        self._jid_label.add_css_class("dimmed")
        box.append(self._jid_label)

        grid.attach(box, 1, 0, 1, 1)

        self.set_child(grid)

    def do_unroot(self) -> None:
        Gtk.ListBoxRow.do_unroot(self)
        app.check_finalize(self)

    def _get_avatar_image(self, contact: types.BareContact | None) -> Gtk.Image:
        if contact is None:
            icon_name = "lucide-user-symbolic"
            return Gtk.Image.new_from_icon_name(icon_name)

        scale = self.get_scale_factor()
        texture = contact.get_avatar(AvatarSize.ROSTER, scale)
        return Gtk.Image.new_from_paintable(texture)

    def update_jid(self, jid: str) -> None:
        self.jid = jid
        self._jid_label.set_text(jid)

    def get_search_text(self):
        if self._contact is None:
            return str(self.jid)
        return f"{self._name} {self.jid}"


class GroupChatInviter(Gtk.Box, SignalManager):
    __gsignals__ = {"listbox-changed": (GObject.SignalFlags.RUN_LAST, None, (bool,))}

    def __init__(self, room_jid: str) -> None:
        Gtk.Box.__init__(self, height_request=250)
        SignalManager.__init__(self)

        self._ui = get_builder("groupchat_inviter.ui")
        self.append(self._ui.invite_box)

        self._invitees_box = ContactsFlowBox()
        self._connect(self._invitees_box, "contact-removed", self._on_invitee_removed)
        self._ui.invitees_scrolled.set_child(self._invitees_box)

        self._ui.contacts_listbox.set_filter_func(self._filter_func, None)
        self._ui.contacts_listbox.set_sort_func(self._sort_func, None)
        self._connect(
            self._ui.contacts_listbox, "row-activated", self._on_contacts_row_activated
        )

        self._room_jid = room_jid

        self._new_contact_rows: dict[str, ContactRow | None] = {}
        self._accounts: list[list[str]] = []

        self._connect(self._ui.search_entry, "search-changed", self._on_search_changed)
        self._connect(
            self._ui.search_entry, "next-match", self._select_new_match, Direction.NEXT
        )
        self._connect(
            self._ui.search_entry,
            "previous-match",
            self._select_new_match,
            Direction.PREV,
        )
        self._connect(self._ui.search_entry, "stop-search", self._on_stop_search)
        self._connect(self._ui.search_entry, "activate", self._on_search_activate)

        controller = Gtk.EventControllerKey()
        self._connect(controller, "key-pressed", self._on_key_pressed)
        self.add_controller(controller)

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.Box.do_unroot(self)

    def _add_accounts(self) -> None:
        for account in self._accounts:
            self._ui.account_store.append([None, *account])

    def _add_contacts(self) -> None:
        show_account = len(self._accounts) > 1
        for account, _label in self._accounts:
            self._new_contact_rows[account] = None
            client = app.get_client(account)
            for contact in client.get_module("Roster").iter_contacts():
                # Exclude group chats
                if contact.is_groupchat:
                    continue
                # Exclude our own jid
                if contact.jid == client.get_own_jid().bare:
                    continue

                row = ContactRow(
                    account, contact, str(contact.jid), contact.name, show_account
                )
                self._ui.contacts_listbox.append(row)

    def _on_contacts_row_activated(self, listbox: Gtk.ListBox, row: ContactRow) -> None:
        self._invitees_box.add_contact(row.account, row.jid, is_new=row.is_new)

        if row.is_new:
            self._remove_new_contact_row()
        else:
            listbox.remove(row)

        self._ui.search_entry.set_text("")
        GLib.timeout_add(50, self._select_first_row)

        invitable = self._invitees_box.has_contacts()
        self.emit("listbox-changed", invitable)

    def _on_invitee_removed(
        self, flowbox: ContactsFlowBox, account: str, jid: str, is_new: bool
    ) -> None:
        if not is_new:
            show_account = len(self._accounts) > 1
            client = app.get_client(account)
            contact = client.get_module("Contacts").get_contact(jid)
            assert isinstance(contact, BareContact)
            row = ContactRow(
                account, contact, str(contact.jid), contact.name, show_account
            )
            self._ui.contacts_listbox.append(row)

        self.emit("listbox-changed", flowbox.has_contacts())

    def _on_key_pressed(
        self,
        _event_controller_key: Gtk.EventControllerKey,
        keyval: int,
        _keycode: int,
        _state: Gdk.ModifierType,
    ) -> bool:
        if keyval == Gdk.KEY_Down:
            self._ui.search_entry.emit("next-match")
            return Gdk.EVENT_STOP

        if keyval == Gdk.KEY_Up:
            self._ui.search_entry.emit("previous-match")
            return Gdk.EVENT_STOP

        if keyval in (
            Gdk.KEY_Return,
            Gdk.KEY_KP_Enter,
            Gdk.KEY_ISO_Enter,
        ):
            row = self._ui.contacts_listbox.get_selected_row()
            if row is not None:
                row.emit("activate")
            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    @staticmethod
    def _on_stop_search(entry: Gtk.SearchEntry) -> None:
        entry.set_text("")

    def _on_search_activate(self, _entry: Gtk.SearchEntry) -> None:
        row = self._ui.contacts_listbox.get_selected_row()
        if row is not None and row.get_child_visible():
            row.emit("activate")

    def _on_search_changed(self, entry: Gtk.SearchEntry) -> None:
        search_text = entry.get_text()
        if "@" in search_text:
            try:
                validate_jid(search_text)
            except Exception:
                self._remove_new_contact_row()
            else:
                self._add_new_contact_row()
                self._update_new_contact_rows(search_text)
        else:
            self._remove_new_contact_row()
        self._ui.contacts_listbox.invalidate_filter()

    def _add_new_contact_row(self) -> None:
        for account, _label in self._accounts:
            row = self._new_contact_rows.get(account)
            if row is not None:
                continue
            show_account = len(self._accounts) > 1
            row = ContactRow(account, None, "", _("New Contact"), show_account)
            self._new_contact_rows[account] = row
            self._ui.contacts_listbox.append(row)

    def _remove_new_contact_row(self) -> None:
        for row in self._new_contact_rows.values():
            if row is not None:
                self._ui.contacts_listbox.remove(row)
        self._new_contact_rows.clear()

    def _update_new_contact_rows(self, search_text: str) -> None:
        for row in self._new_contact_rows.values():
            if row is not None:
                row.update_jid(search_text)

    def _select_new_match(self, _entry: Gtk.SearchEntry, direction: Direction) -> None:
        selected_row = self._ui.contacts_listbox.get_selected_row()
        if selected_row is None:
            return

        index = selected_row.get_index()

        if direction == Direction.NEXT:
            index += 1
        else:
            index -= 1

        while True:
            new_selected_row = self._ui.contacts_listbox.get_row_at_index(index)
            if new_selected_row is None:
                return
            if new_selected_row.get_child_visible():
                self._ui.contacts_listbox.select_row(new_selected_row)
                return

            if direction == Direction.NEXT:
                index += 1
            else:
                index -= 1

    def _select_first_row(self) -> None:
        first_row = self._ui.contacts_listbox.get_row_at_y(0)
        self._ui.contacts_listbox.select_row(first_row)

    def _scroll_to_first_row(self) -> None:
        self._ui.scrolledwindow.get_vadjustment().set_value(0)

    def _filter_func(self, row: ContactRow, _user_data: Any | None) -> bool:
        search_text = self._ui.search_entry.get_text().lower()
        search_text_list = search_text.split()
        row_text = row.get_search_text().lower()
        for text in search_text_list:
            if text not in row_text:
                GLib.timeout_add(50, self._select_first_row)
                return False
        GLib.timeout_add(50, self._select_first_row)
        return True

    @staticmethod
    def _sort_func(row1: ContactRow, row2: ContactRow, _user_data: Any | None) -> int:
        name1 = row1.get_search_text()
        name2 = row2.get_search_text()
        account1 = row1.account
        account2 = row2.account

        result = locale.strcoll(account1.lower(), account2.lower())
        if result != 0:
            return result

        return locale.strcoll(name1.lower(), name2.lower())

    def _reset(self) -> None:
        for row in iterate_listbox_children(self._ui.contacts_listbox):
            self._ui.contacts_listbox.remove(row)

        self._invitees_box.clear()

    def load_contacts(self) -> None:
        self._reset()

        self._accounts = app.get_enabled_accounts_with_labels()
        self._new_contact_rows = {}
        self._add_accounts()

        self._add_contacts()

        first_row = self._ui.contacts_listbox.get_row_at_index(0)
        self._ui.contacts_listbox.select_row(first_row)
        self._ui.search_entry.grab_focus()
        self.emit("listbox-changed", False)

    def get_invitees(self) -> list[str]:
        return self._invitees_box.get_contact_jids()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from functools import lru_cache

from gi.repository import Gdk
from gi.repository import GtkSource
from nbxmpp import util as nbxmpp_util

from gajim.common import app
from gajim.common import types
from gajim.common.modules.contacts import GroupchatOfflineParticipant
from gajim.common.modules.contacts import GroupchatParticipant


def text_to_color(text: str) -> tuple[float, float, float]:
    if app.css_config.prefer_dark:
        lightness = 60
    else:
        lightness = 40
    return nbxmpp_util.text_to_color(text, 100, lightness)


def get_contact_color(
    contact: types.ChatContactT | GroupchatOfflineParticipant,
) -> tuple[float, float, float]:
    if isinstance(contact, GroupchatParticipant | GroupchatOfflineParticipant):
        if contact.room.muc_context in (None, "public"):
            return text_to_color(contact.name)

        if contact.real_jid is not None:
            return text_to_color(str(contact.real_jid))

    return text_to_color(str(contact.jid))


def get_color_for_account(account: str) -> str:
    col_r, col_g, col_b = text_to_color(account)
    rgba = Gdk.RGBA()
    rgba.red = col_r
    rgba.green = col_g
    rgba.blue = col_b
    rgba.alpha = 1
    return rgba.to_string()


def convert_rgba_to_hex(rgba: Gdk.RGBA) -> str:
    red = int(rgba.red * 255)
    green = int(rgba.green * 255)
    blue = int(rgba.blue * 255)
    return f"#{red:02x}{green:02x}{blue:02x}"


@lru_cache(maxsize=1024)
def convert_rgb_string_to_float(rgb_string: str) -> tuple[float, float, float]:
    rgba = Gdk.RGBA()
    rgba.parse(rgb_string)
    return (rgba.red, rgba.green, rgba.blue)


def rgba_to_float(rgba: Gdk.RGBA) -> tuple[float, float, float]:
    return (rgba.red, rgba.green, rgba.blue)


def make_rgba(color_string: str) -> Gdk.RGBA:
    rgba = Gdk.RGBA()
    rgba.parse(color_string)
    return rgba


@lru_cache(maxsize=16)
def get_css_show_class(show: str) -> str:
    if show in ("online", "chat"):
        return ".gajim-status-online"
    if show == "away":
        return ".gajim-status-away"
    if show in ("dnd", "xa"):
        return ".gajim-status-dnd"
    if show == "connecting":
        return ".gajim-status-connecting"
    return ".gajim-status-offline"


def get_source_view_style_scheme() -> GtkSource.StyleScheme | None:
    style_scheme_manager = GtkSource.StyleSchemeManager.get_default()
    if app.css_config.prefer_dark:
        return style_scheme_manager.get_scheme("solarized-dark")
    return style_scheme_manager.get_scheme("solarized-light")



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Any
from typing import Literal
from typing import TypedDict

import uuid

from nbxmpp.protocol import JID

from gajim.common.i18n import _


class _DEFAULT:
    pass


class _ACCOUNTDEFAULT:
    pass


HAS_APP_DEFAULT = _DEFAULT()
HAS_ACCOUNT_DEFAULT = _ACCOUNTDEFAULT()


BoolSettings = Literal[
    "always_english_wikipedia",
    "always_english_wiktionary",
    "ask_online_status",
    "autoaway",
    "autoxa",
    "chat_merge_consecutive_nickname",
    "check_for_update",
    "confirm_close_muc",
    "confirm_on_window_delete",
    "dev_force_bookmark_2",
    "enable_emoji_shortcodes",
    "enable_keepassxc_integration",
    "escape_key_closes",
    "enable_link_preview_default",
    "gc_enable_link_preview_default",
    "gc_notify_on_all_messages_private_default",
    "gc_notify_on_all_messages_public_default",
    "gc_notify_on_reaction_default",
    "gc_print_join_left_default",
    "gc_print_status_default",
    "hide_groupchat_occupants_list",
    "ignore_incoming_attention",
    "is_window_visible",
    "muc_prefer_direct_msg",
    "notification_preview_message",
    "notify_on_all_muc_messages",
    "notify_on_file_complete",
    "notify_on_reaction_default",
    "plugins_auto_update",
    "plugins_notify_after_update",
    "plugins_repository_enabled",
    "plugins_update_check",
    "positive_184_ack",
    "preview_allow_all_images",
    "preview_anonymous_muc",
    "preview_verify_https",
    "print_status_in_chats",
    "remote_control",
    "send_on_ctrl_enter",
    "show_help_start_chat",
    "show_in_taskbar",
    "show_notifications",
    "show_notifications_away",
    "show_send_message_button",
    "show_voice_message_button",
    "show_status_msgs_in_roster",
    "show_subject_on_join",
    "show_trayicon",
    "sort_by_show_in_muc",
    "sort_by_show_in_start_chat",
    "sounddnd",
    "sounds_on",
    "trayicon_notification_on_events",
    "use_http2",
    "use_keyring",
    "use_kib_mib",
    "use_speller",
    "use_stun_server",
    "use_urgency_hint",
    "video_see_self",
    "show_header_bar",
]

IntSettings = Literal[
    "audio_input_volume",
    "audio_output_volume",
    "autoawaytime",
    "autoxatime",
    "chat_handle_position",
    "dark_theme",
    "file_transfers_port",
    "gc_sync_threshold_private_default",
    "gc_sync_threshold_public_default",
    "groupchat_roster_width",
    "mainwin_height",
    "mainwin_width",
    "preview_max_file_size",
    "preview_size",
]

FloatSettings = Literal["app_font_size",]

StringSettings = Literal[
    "action_on_close",
    "additional_uri_schemes",
    "audio_input_device",
    "audio_output_device",
    "autoaway_message",
    "autoxa_message",
    "confirm_block",
    "date_format",
    "date_time_format",
    "dictionary_url",
    "gc_refer_to_nick_char",
    "global_proxy",
    "last_save_dir",
    "last_send_dir",
    "last_sounds_dir",
    "last_update_check",
    "latest_disco_addresses",
    "muc_highlight_words",
    "muclumbus_api_http_uri",
    "muclumbus_api_jid",
    "muclumbus_api_pref",
    "providers_list_url",
    "quick_reaction_emojis",
    "roster_theme",
    "search_engine",
    "show_main_window_on_startup",
    "speller_language",
    "stun_server",
    "time_format",
    "video_framerate",
    "video_input_device",
    "video_size",
]

AllSettings = Literal[BoolSettings, IntSettings, FloatSettings, StringSettings]
AllSettingsT = str | int | float | bool | list[str]

APP_SETTINGS: dict[str, str | int | float | bool | list[Any]] = {
    "additional_uri_schemes": "",
    "always_english_wikipedia": False,
    "always_english_wiktionary": True,
    "app_font_size": 1.0,
    "ask_online_status": False,
    "audio_input_device": "autoaudiosrc ! volume name=gajim_vol",
    "audio_input_volume": 50,
    "audio_output_device": "autoaudiosink",
    "audio_output_volume": 50,
    "autoaway": True,
    "autoaway_message": "",
    "autoawaytime": 5,
    "autoxa": True,
    "autoxa_message": "",
    "autoxatime": 15,
    "chat_handle_position": 350,
    "chat_merge_consecutive_nickname": True,
    "check_for_update": True,
    "confirm_block": "",
    "confirm_close_muc": True,
    "confirm_on_window_delete": True,
    "dark_theme": 2,
    "date_format": "%x",
    "date_time_format": "%c",
    "dev_force_bookmark_2": False,
    "dictionary_url": "WIKTIONARY",
    "enable_emoji_shortcodes": True,
    "enable_keepassxc_integration": False,
    "escape_key_closes": False,
    "enable_link_preview_default": True,
    "file_transfers_port": 28011,
    "ft_add_hosts_to_send": "",
    "gc_enable_link_preview_default": True,
    "gc_notify_on_all_messages_private_default": True,
    "gc_notify_on_all_messages_public_default": False,
    "gc_notify_on_reaction_default": True,
    "gc_print_join_left_default": False,
    "gc_print_status_default": False,
    "gc_refer_to_nick_char": ",",
    "gc_sync_threshold_private_default": 0,
    "gc_sync_threshold_public_default": 1,
    "global_proxy": "",
    "groupchat_roster_width": 250,
    "hide_groupchat_occupants_list": False,
    "ignore_incoming_attention": False,
    "is_window_visible": True,
    "last_save_dir": "",
    "last_send_dir": "",
    "last_sounds_dir": "",
    "last_update_check": "",
    "latest_disco_addresses": "",
    "mainwin_height": 500,
    "mainwin_width": 1000,
    "action_on_close": "hide",
    "muc_highlight_words": "",
    "muc_prefer_direct_msg": True,
    "muclumbus_api_http_uri": "https://search.jabber.network/api/1.0/search",
    "muclumbus_api_jid": "api@search.jabber.network",
    "muclumbus_api_pref": "http",
    "notification_preview_message": True,
    "notify_on_all_muc_messages": False,
    "notify_on_file_complete": True,
    "notify_on_reaction_default": True,
    "plugins_auto_update": False,
    "plugins_notify_after_update": True,
    "plugins_repository_enabled": True,
    "plugins_update_check": True,
    "positive_184_ack": False,
    "preview_allow_all_images": False,
    "preview_anonymous_muc": False,
    "preview_max_file_size": 10485760,
    "preview_size": 400,
    "preview_verify_https": True,
    "providers_list_url": "https://data.xmpp.net/providers/v2/providers-B.json",
    "print_status_in_chats": False,
    "quick_reaction_emojis": "👍,❤,🤣",
    "remote_control": False,
    "roster_theme": "default",
    "search_engine": "https://duckduckgo.com/?q=%s",
    "send_on_ctrl_enter": False,
    "show_header_bar": True,
    "show_help_start_chat": True,
    "show_in_taskbar": True,
    "show_main_window_on_startup": "always",
    "show_notifications": True,
    "show_notifications_away": False,
    "show_send_message_button": False,
    "show_voice_message_button": True,
    "show_status_msgs_in_roster": True,
    "show_subject_on_join": True,
    "show_trayicon": True,
    "sort_by_show_in_muc": False,
    "sort_by_show_in_start_chat": True,
    "sounddnd": False,
    "sounds_on": True,
    "speller_language": "",
    "stun_server": "",
    "time_format": "%H:%M",
    "trayicon_notification_on_events": True,
    "use_http2": False,
    "use_keyring": True,
    "use_kib_mib": False,
    "use_speller": True,
    "use_stun_server": False,
    "use_urgency_hint": True,
    "video_framerate": "",
    "video_input_device": "autovideosrc",
    "video_see_self": True,
    "video_size": "",
    "workspace_order": [],
}

BoolAccountSettings = Literal[
    "active",
    "anonymous_auth",
    "answer_receipts",
    "autoauth",
    "autoconnect",
    "autojoin_sync",
    "client_cert_encrypted",
    "confirm_unencrypted_connection",
    "enable_gssapi",
    "enable_security_labels",
    "ft_send_local_ips",
    "gc_send_marker_private_default",
    "gc_send_marker_public_default",
    "ignore_unknown_contacts",
    "publish_location",
    "publish_tune",
    "request_user_data",
    "restore_last_status",
    "savepass",
    "send_idle_time",
    "send_marker_default",
    "send_os_info",
    "sync_muc_blocks",
    "sync_with_global_status",
    "test_ft_proxies_on_startup",
    "update_timezone",
    "use_custom_host",
    "use_ft_proxies",
    "use_plain_connection",
    "omemo_blind_trust",
    "openpgp_blind_trust",
    "openpgp_backup_secret_key",
]


StringAccountSettings = Literal[
    "account_color",
    "account_label",
    "address",
    "avatar_sha",
    "client_cert",
    "custom_host",
    "custom_type",
    "default_workspace",
    "file_transfer_proxies",
    "filetransfer_preference",
    "gc_send_chatstate_default",
    "http_auth",
    "last_status",
    "last_status_msg",
    "password",
    "proxy",
    "resource",
    "roster_version",
    "send_chatstate_default",
    "subscription_request_msg",
]

IntAccountSettings = Literal[
    "chat_history_max_age",
    "custom_port",
]


AllAccountSettings = Literal[
    BoolAccountSettings, IntAccountSettings, StringAccountSettings
]

AllAccountSettingsT = bool | int | str


BoolGroupChatSettings = Literal[
    "notify_on_all_messages",
    "print_join_left",
    "print_status",
    "send_marker",
    "enable_link_preview",
]

StringGroupChatSettings = Literal[
    "encryption",
    "mute_until",
    "speller_language",
    "send_chatstate",
]

IntGroupChatSettings = Literal["sync_threshold",]


AllGroupChatSettings = Literal[
    BoolGroupChatSettings, IntGroupChatSettings, StringGroupChatSettings
]

AllGroupChatSettingsT = str | int | bool


BoolContactSettings = Literal[
    "send_marker",
    "enable_link_preview",
]

StringContactSettings = Literal[
    "encryption",
    "mute_until",
    "speller_language",
    "send_chatstate",
]

AllContactSettings = Literal[BoolContactSettings, StringContactSettings]

AllContactSettingsT = str | bool


ACCOUNT_SETTINGS = {
    "account": {
        "account_color": "rgb(85, 85, 85)",
        "account_label": "",
        "active": False,
        "address": "",
        "anonymous_auth": False,
        "answer_receipts": True,
        "autoauth": False,
        "autoconnect": True,
        "autojoin_sync": True,
        "avatar_sha": "",
        "chat_history_max_age": -1,
        "client_cert": "",
        "client_cert_encrypted": False,
        "confirm_unencrypted_connection": True,
        "custom_host": "",
        "custom_port": 5222,
        "custom_type": "START TLS",
        "default_workspace": "",
        "enable_gssapi": False,
        "enable_security_labels": False,
        "encryption_default": "",
        "file_transfer_proxies": "",
        "filetransfer_preference": "httpupload",
        "ft_send_local_ips": True,
        "gc_send_chatstate_default": "composing_only",
        "gc_send_marker_private_default": True,
        "gc_send_marker_public_default": False,
        "http_auth": "ask",
        "ignore_unknown_contacts": False,
        "last_status": "online",
        "last_status_msg": "",
        "password": "",
        "proxy": "",
        "publish_location": False,
        "publish_tune": False,
        "request_user_data": True,
        "resource": "gajim.$rand",
        "restore_last_status": False,
        "roster_version": "",
        "savepass": True,
        "send_chatstate_default": "composing_only",
        "send_idle_time": True,
        "send_marker_default": True,
        "send_os_info": True,
        "subscription_request_msg": "",
        "sync_muc_blocks": True,
        "sync_with_global_status": True,
        "test_ft_proxies_on_startup": False,
        "update_timezone": True,
        "use_custom_host": False,
        "use_ft_proxies": False,
        "use_plain_connection": False,
        "omemo_blind_trust": True,
        "openpgp_blind_trust": True,
        "openpgp_backup_secret_key": True,
    },
    "contact": {
        "enable_link_preview": HAS_APP_DEFAULT,
        "encryption": HAS_ACCOUNT_DEFAULT,
        "mute_until": "",
        "send_chatstate": HAS_ACCOUNT_DEFAULT,
        "send_marker": HAS_ACCOUNT_DEFAULT,
        "speller_language": "",
    },
    "group_chat": {
        "enable_link_preview": HAS_APP_DEFAULT,
        "encryption": HAS_ACCOUNT_DEFAULT,
        "mute_until": "",
        "notify_on_all_messages": HAS_APP_DEFAULT,
        "print_join_left": HAS_APP_DEFAULT,
        "print_status": HAS_APP_DEFAULT,
        "send_chatstate": HAS_ACCOUNT_DEFAULT,
        "send_marker": HAS_ACCOUNT_DEFAULT,
        "speller_language": "",
        "sync_threshold": HAS_APP_DEFAULT,
    },
}

StringWorkspaceSettings = Literal[
    "avatar_sha",
    "color",
    "name",
]

AllWorkspaceSettings = Literal[StringWorkspaceSettings, "chats"]


class OpenChatSettingDetails(TypedDict):
    account: str
    jid: JID
    type: Literal["chat", "groupchat", "pm"]
    pinned: bool
    position: int


OpenChatsSettingT = list[OpenChatSettingDetails]
AllWorkspaceSettingsT = str | OpenChatsSettingT


class WorkspaceSettings(TypedDict):
    name: str
    color: str
    avatar_sha: str
    chats: OpenChatsSettingT


WORKSPACE_SETTINGS: WorkspaceSettings = {
    "name": _("My Workspace"),
    "color": "",
    "avatar_sha": "",
    "chats": [],
}


INITAL_WORKSPACE: dict[str, dict[str, WorkspaceSettings]] = {str(uuid.uuid4()): {}}


PLUGIN_SETTINGS = {"active": False}


PROXY_SETTINGS = {
    "type": "socks5",
    "host": "",
    "port": 0,
    "useauth": False,
    "user": "",
    "pass": "",
}


PROXY_EXAMPLES = {
    "Tor": {"type": "socks5", "host": "localhost", "port": 9050},
    "Tor (Browser)": {"type": "socks5", "host": "localhost", "port": 9150},
}


DEFAULT_SOUNDEVENT_SETTINGS = {
    "attention_received": {"enabled": True, "path": "attention.wav"},
    "first_message_received": {"enabled": True, "path": "message1.wav"},
    "contact_connected": {"enabled": False, "path": "connected.wav"},
    "contact_disconnected": {"enabled": False, "path": "disconnected.wav"},
    "message_sent": {"enabled": False, "path": "sent.wav"},
    "muc_message_highlight": {"enabled": True, "path": "gc_message1.wav"},
    "muc_message_received": {"enabled": True, "path": "message2.wav"},
    "incoming-call-sound": {"enabled": True, "path": "call_incoming.wav"},
    "outgoing-call-sound": {"enabled": True, "path": "call_outgoing.wav"},
}


ADVANCED_SETTINGS = {
    "app": {
        "additional_uri_schemes": _(
            "Clickable schemes in addition to the hard-coded list of "
            "IANA-registered ones. Space-separated, lower-case, no colons."
        ),
        "always_english_wikipedia": "",
        "always_english_wiktionary": "",
        "chat_merge_consecutive_nickname": _(
            "Show message meta data (avatar, nickname, timestamp) only once, "
            "if there are multiple messages from the same sender within a "
            "specific timespan."
        ),
        "confirm_block": _(
            "Show a confirmation dialog to block a contact? Empty string "
            "means never show the dialog."
        ),
        "confirm_close_muc": _("Ask before closing a group chat tab/window."),
        "confirm_on_window_delete": _(
            "Ask before quitting when Gajim’s window is closed"
        ),
        "date_format": "https://docs.python.org/3/library/time.html#time.strftime",  # noqa: E501
        "date_time_format": "https://docs.python.org/3/library/time.html#time.strftime",  # noqa: E501
        "dev_force_bookmark_2": _("Force Bookmark 2 usage"),
        "dictionary_url": _(
            "Either a custom URL with %%s in it (where %%s is the word/phrase)"
            ' or "WIKTIONARY" (which means use Wikitionary).'
        ),
        "file_transfers_port": "",
        "ft_add_hosts_to_send": _(
            "List of send hosts (comma separated) in "
            "addition to local interfaces for file transfers (in case of "
            "address translation/port forwarding)."
        ),
        "gc_notify_on_all_messages_private_default": "",
        "gc_notify_on_all_messages_public_default": "",
        "gc_refer_to_nick_char": _(
            "Character to add after nickname when using nickname completion "
            "(tab) in group chat."
        ),
        "groupchat_roster_width": _("Width of group chat roster in pixel"),
        "ignore_incoming_attention": _(
            'If enabled, Gajim will ignore incoming attention requests ("wizz").'
        ),
        "muc_highlight_words": _(
            "A list of words (semicolon separated) that will be "
            "highlighted in group chats."
        ),
        "muclumbus_api_http_uri": "",
        "muclumbus_api_jid": "",
        "muclumbus_api_pref": _('API Preferences. Possible values: "http", "iq"'),
        "notification_preview_message": _(
            "Preview new messages in notification popup?"
        ),
        "notify_on_all_muc_messages": "",
        "plugins_repository_enabled": _(
            "If enabled, Gajim offers to download plugins hosted on gajim.org"
        ),
        "providers_list_url": _(
            "Endpoint for retrieving a list of providers for sign up"
        ),
        "quick_reaction_emojis": _(
            "Comma-separated list of emojis for quick reaction buttons"
        ),
        "search_engine": "",
        "stun_server": _("STUN server to use when using Jingle"),
        "time_format": "https://docs.python.org/3/library/time.html#time.strftime",  # noqa: E501
        "use_http2": _("Use HTTP2 for requests if there is server support"),
        "use_kib_mib": _("IEC standard says KiB = 1024 bytes, KB = 1000 bytes."),
        "use_stun_server": _(
            "If enabled, Gajim will try to use a STUN server when using Jingle."
            ' The one in "stun_server" option, or the one given by '
            "the XMPP server."
        ),
        "use_urgency_hint": _(
            "If enabled, Gajim makes the window flash (the default behaviour "
            "in most Window Managers) when holding pending events."
        ),
    },
}



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0092: Software Version

from __future__ import annotations

from nbxmpp.protocol import JID

from gajim.common import app
from gajim.common import types
from gajim.common.modules.base import BaseModule
from gajim.common.modules.contacts import BareContact
from gajim.common.util.version import get_os_info


class SoftwareVersion(BaseModule):
    _nbxmpp_extends = "SoftwareVersion"
    _nbxmpp_methods = [
        "set_software_version",
        "request_software_version",
        "disable",
    ]

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

    def set_enabled(self, enabled: bool) -> None:
        if enabled and app.settings.get_account_setting(self._account, "send_os_info"):
            os_info = get_os_info()
        else:
            os_info = None
        self._nbxmpp("SoftwareVersion").set_software_version(
            "Gajim", app.version, os_info
        )
        self._nbxmpp("SoftwareVersion").set_allow_reply_func(self._allow_reply)

    def _allow_reply(self, jid: JID) -> bool:
        item = self._con.get_module("Roster").get_item(jid.new_as_bare())
        if item is None:
            return False

        contact = self._get_contact(JID.from_string(jid.bare))
        assert isinstance(contact, BareContact)
        return contact.is_subscribed



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import logging
import time

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import Gtk
from nbxmpp import JID
from nbxmpp.namespaces import Namespace
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import MucSubject

from gajim.common import app
from gajim.common.const import AvatarSize
from gajim.common.const import RFC5646_LANGUAGE_TAGS
from gajim.common.const import XmppUriQuery
from gajim.common.i18n import _
from gajim.common.i18n import p_
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.util.muc import get_groupchat_name
from gajim.common.util.text import make_href_markup

from gajim.gtk.builder import get_builder
from gajim.gtk.contact_name_entry import ContactNameEntry
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import clear_listbox
from gajim.gtk.util.misc import container_remove_all
from gajim.gtk.util.misc import open_uri

log = logging.getLogger("gajim.gtk.groupchat_info")

MUC_FEATURES = {
    "muc_open": (
        "lucide-globe-symbolic",
        p_("Group chat feature", "Open"),
        _("Anyone can join this group chat"),
    ),
    "muc_membersonly": (
        "lucide-user-check-symbolic",
        p_("Group chat feature", "Members Only"),
        _("This group chat is restricted to members only"),
    ),
    "muc_nonanonymous": (
        "lucide-eye-symbolic",
        p_("Group chat feature", "Not Anonymous"),
        _("All other group chat participants can see your XMPP address"),
    ),
    "muc_semianonymous": (
        "lucide-eye-off-symbolic",
        p_("Group chat feature", "Semi-Anonymous"),
        _("Only moderators can see your XMPP address"),
    ),
    "muc_moderated": (
        "lucide-mic-off-symbolic",
        p_("Group chat feature", "Moderated"),
        _(
            "Participants entering this group chat need "
            "to request permission to send messages"
        ),
    ),
    "muc_unmoderated": (
        "lucide-mic-symbolic",
        p_("Group chat feature", "Not Moderated"),
        _("Participants entering this group chat are allowed to send messages"),
    ),
    "muc_public": (
        "lucide-megaphone-symbolic",
        p_("Group chat feature", "Public"),
        _("Group chat can be found via search"),
    ),
    "muc_hidden": (
        "lucide-megaphone-off-symbolic",
        p_("Group chat feature", "Hidden"),
        _("This group chat can not be found via search"),
    ),
    "muc_passwordprotected": (
        "lucide-lock-symbolic",
        p_("Group chat feature", "Password Required"),
        _("This group chat does require a password upon entry"),
    ),
    "muc_unsecured": (
        "lucide-lock-open-symbolic",
        p_("Group chat feature", "No Password Required"),
        _("This group chat does not require a password upon entry"),
    ),
    "muc_persistent": (
        "lucide-hard-drive-symbolic",
        p_("Group chat feature", "Persistent"),
        _("This group chat persists even if there are no participants"),
    ),
    "muc_temporary": (
        "lucide-clock-symbolic",
        p_("Group chat feature", "Temporary"),
        _("This group chat will be destroyed once the last participant left"),
    ),
    "mam": (
        "lucide-server-symbolic",
        p_("Group chat feature", "Archiving"),
        _("Messages are archived on the server"),
    ),
}


class GroupChatInfoScrolled(Gtk.ScrolledWindow, SignalManager):
    __gsignals__ = {
        "name-updated": (GObject.SignalFlags.RUN_LAST, None, (str,)),
    }

    def __init__(
        self,
        account: str | None = None,
        width: int = 300,
        minimal: bool = False,
        edit_mode: bool = False,
        show_users: bool = True,
        add_padding: bool = True,
    ) -> None:
        SignalManager.__init__(self)
        Gtk.ScrolledWindow.__init__(self, width_request=width)

        self._minimal = minimal
        self._show_users = show_users

        if minimal:
            self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.NEVER)
        else:
            self.set_vexpand(True)
            self.set_hexpand(True)
            self.set_propagate_natural_width(True)
            self.set_min_content_height(400)
            self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)

        self._account = account
        self._contact: GroupchatContact | None = None
        self._info: DiscoInfo | None = None

        self._ui = get_builder("groupchat_info_scrolled.ui")
        self.set_child(self._ui.info_clamp)

        if add_padding:
            self._ui.info_clamp.add_css_class("p-3")

        self._connect(self._ui.address_copy_button, "clicked", self._on_copy_address)

        self._contact_name_entry = ContactNameEntry(editable=edit_mode)
        self._connect(
            self._contact_name_entry, "name-updated", self._on_contact_name_updated
        )

        self._ui.name_box.append(self._contact_name_entry)

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.ScrolledWindow.do_unroot(self)
        del self._contact_name_entry
        app.check_finalize(self)

    def get_account(self) -> str | None:
        return self._account

    def set_account(self, account: str) -> None:
        self._account = account

    def get_jid(self) -> JID | None:
        if self._contact is not None:
            return self._contact.jid
        if self._info is not None:
            return self._info.jid
        return None

    def set_subject(self, muc_subject: MucSubject | None) -> None:
        if muc_subject is None:
            return

        self._ui.subject_row.set_subtitle(make_href_markup(muc_subject.text))
        self._ui.subject_row.set_visible(bool(muc_subject.text))

        author = muc_subject.author
        has_author = bool(author)
        if has_author and muc_subject.timestamp is not None:
            time_ = time.strftime("%c", time.localtime(muc_subject.timestamp))
            author = f"{muc_subject.author} - {time_}"

        self._ui.author_row.set_subtitle(author or "")
        self._ui.author_row.set_visible(has_author)

    def set_info_from_contact(self, contact: GroupchatContact) -> None:
        self._contact = contact

        self._contact_name_entry.set_contact(contact)

        disco_info = contact.get_disco()
        if disco_info is not None:
            self.set_from_disco_info(disco_info)
            return

        self._ui.address_row.set_subtitle(str(contact.jid))
        self._ui.address_copy_button.set_sensitive(True)

    def set_from_disco_info(self, info: DiscoInfo) -> None:
        self._info = info
        # Set name
        if self._account is None:
            name = info.muc_name
        else:
            client = app.get_client(self._account)
            assert info.jid is not None
            name = get_groupchat_name(client, info.jid)
            contact = client.get_module("Contacts").get_contact(
                info.jid, groupchat=True
            )
            assert isinstance(contact, GroupchatContact)
            texture = contact.get_avatar(AvatarSize.GROUP_INFO, self.get_scale_factor())
            self._ui.avatar_image.set_pixel_size(AvatarSize.GROUP_INFO)
            self._ui.avatar_image.set_from_paintable(texture)

        self._contact_name_entry.update_displayed_name(name or "")

        self._ui.address_row.set_subtitle(str(info.jid))
        self._ui.address_copy_button.set_sensitive(True)

        self._ui.description_row.set_subtitle(make_href_markup(info.muc_description))
        self._ui.description_row.set_visible(bool(info.muc_description))

        if self._minimal:
            return

        self._ui.features_group.set_visible(True)
        self._ui.users_row.set_subtitle(info.muc_users or "")
        self._ui.users_row.set_visible(info.muc_users is not None and self._show_users)

        # Set contacts
        container_remove_all(self._ui.contact_box)
        if info.muc_contacts:
            for contact in info.muc_contacts:
                try:
                    jid = JID.from_string(contact).new_as_bare()
                except Exception as e:
                    log.debug("Bad MUC contact address %s: %s", contact, str(e))
                else:
                    self._ui.contact_box.append(self._get_contact_button(jid))

        self._ui.contact_row.set_visible(bool(info.muc_contacts))

        # Message history
        self._ui.logs_row.set_visible(bool(info.muc_log_uri))
        self._ui.logs.set_uri(info.muc_log_uri or "")

        # Set room language
        lang = ""
        if info.muc_lang:
            lang = RFC5646_LANGUAGE_TAGS.get(info.muc_lang, info.muc_lang)
        self._ui.lang_row.set_subtitle(lang or "")
        self._ui.lang_row.set_visible(bool(info.muc_lang))

        self._add_features(info.features)

    def enable_edit_mode(self) -> None:
        self._contact_name_entry.enable_edit_mode()

    def _on_contact_name_updated(self, _widget: ContactNameEntry, name: str) -> None:
        self.emit("name-updated", name)

    def _add_features(self, features: list[str]) -> None:
        features = list(features)

        if Namespace.MAM_2 in features:
            features.append("mam")

        clear_listbox(self._ui.features_listbox)
        for feature in MUC_FEATURES:
            if feature in features:
                icon, title, subtitle = MUC_FEATURES.get(feature, (None, None, None))
                if icon is None or title is None or subtitle is None:
                    log.debug("No user strings for feature %s", feature)
                    continue

                row = Adw.ActionRow(title=title, subtitle=subtitle)
                image = Gtk.Image.new_from_icon_name(icon)
                image.set_pixel_size(24)
                image.add_css_class("dimmed")
                row.add_prefix(image)

                self._ui.features_listbox.append(row)

    def _on_copy_address(self, _button: Gtk.Button) -> None:
        jid = None
        if self._contact is not None:
            jid = self._contact.jid
        else:
            if self._info is not None:
                jid = self._info.jid

        if jid is None:
            return

        self.get_clipboard().set(jid.to_iri(XmppUriQuery.JOIN.value))

    def _on_activate_contact_link(self, button: Gtk.LinkButton) -> int:
        open_uri(button.get_uri())
        return Gdk.EVENT_STOP

    def _get_contact_button(self, contact: JID) -> Gtk.Button:
        button = Gtk.LinkButton(
            uri=contact.to_iri(XmppUriQuery.MESSAGE.value), label=str(contact)
        )
        button.add_css_class("link-button")
        self._connect(button, "activate-link", self._on_activate_contact_link)
        return button



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import logging
import sys
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path

from gi.repository import GLib  # noqa: E402
from gi.repository import GObject  # noqa: E402
from gi.repository import Gst  # noqa: E402

from gajim.common.enum import AudioPlayerState
from gajim.common.ged import EventHelper

from gajim.gtk.util.classes import SignalManager

AudioSampleT = list[tuple[float, float]]

log = logging.getLogger("gajim.gtk.preview.audio_player")


@dataclass
class AudioPreviewState:
    duration: float = 0.0
    position: float = 0.0
    pipeline_state: AudioPlayerState = AudioPlayerState.NULL
    speed: float = 1.0
    is_timestamp_positive: bool = True
    samples: AudioSampleT = field(default_factory=list[tuple[float, float]])
    is_audio_analyzed = False


class AudioPlayer(GObject.GObject, SignalManager, EventHelper):
    __gtype_name__ = "AudioPlayer"
    __gsignals__ = {
        "audio-playback-changed": (GObject.SignalFlags.RUN_LAST, None, (int, int)),
        "audio-playback-progressed": (GObject.SignalFlags.RUN_LAST, None, (int, int)),
    }

    def __init__(self) -> None:
        super().__init__()
        SignalManager.__init__(self)
        EventHelper.__init__(self)

        self._file_path = None
        self._state = None
        self._preview_id = -1

        self._speed_min = 0.25
        self._speed_max = 2.00

        self._playbin = Gst.ElementFactory.make("playbin", "bin")
        self._query = Gst.Query.new_position(Gst.Format.TIME)
        self._bus_watch_id: int = 0

        self._do_broadcasting = False
        self._loop = False
        self._switch_track = False
        self._play_new_track = False
        self._seek_track_position = False
        self._setup_pipeline()

        # Holds active audio preview sessions
        # for resuming after switching chats
        self._audio_sessions: dict[int, AudioPreviewState] = {}

    def destroy(self) -> None:
        if self._playbin is not None:
            self._playbin.set_state(Gst.State.NULL)
            bus = self._playbin.get_bus()

            if bus is not None:
                bus.remove_signal_watch()
                bus.disconnect(self._bus_watch_id)

    @property
    def preview_id(self) -> int:
        return self._preview_id

    def get_audio_state(self, preview_id: int) -> AudioPreviewState:
        state = self._audio_sessions.get(preview_id)
        if state is not None:
            return state
        self._audio_sessions[preview_id] = AudioPreviewState()
        return self._audio_sessions[preview_id]

    def toggle_playback(self) -> None:
        assert self._playbin is not None

        if self._is_playing():
            self._playbin.set_state(Gst.State.PAUSED)
        else:
            if self._is_eos():
                # Treat re-playing same as playing a new track
                self._switch_track = True
                self._seek_track_position = True
                self._play_new_track = True
                self._set_playback_position(0.0)
            else:
                self._playbin.set_state(Gst.State.PLAYING)

    def play_audio_file(
        self,
        file_path: Path,
        preview_id: int,
        loop: bool = False,
        from_start: bool = False,
    ) -> None:
        assert self._playbin is not None
        assert preview_id != self._preview_id

        self._do_broadcasting = False
        self._loop = loop
        self._playbin.set_state(Gst.State.READY)

        self._preview_id = preview_id

        state = self._audio_sessions.get(preview_id)
        assert state is not None
        self._state = state
        self._file_path = file_path

        self._switch_track = True
        self._seek_track_position = True
        self._play_new_track = True

        self._playbin.set_property("uri", self._file_path.as_uri())

        if self._is_eos() or from_start:
            self._state.position = 0.0
        self._playbin.set_state(Gst.State.PLAYING)

    def pause(self) -> None:
        assert self._playbin is not None
        self._playbin.set_state(Gst.State.PAUSED)

    def stop(self, preview_id: int) -> None:
        assert self._playbin is not None

        if self._preview_id != preview_id:
            return

        self.stop_any()

    def stop_any(self):
        assert self._playbin is not None

        self._do_broadcasting = False
        self._loop = False
        self._preview_id = -1
        self._playbin.set_state(Gst.State.NULL)
        if self._state is not None:
            self._state.pipeline_state = AudioPlayerState.NULL
        self.emit(
            "audio-playback-changed",
            -1,
            AudioPlayerState.NULL.value,
        )

    def set_playback_position(self, preview_id: int, position: float) -> None:
        assert self._playbin is not None

        if preview_id != self._preview_id or self._state is None:
            return

        self._set_playback_position(position)

    def set_playback_speed(self, speed: float) -> bool:
        assert self._state is not None

        if self._speed_min <= speed <= self._speed_max:
            self._state.speed = speed

        assert self._playbin is not None
        self._playbin.seek(
            self._state.speed,
            Gst.Format.TIME,
            Gst.SeekFlags.FLUSH,
            Gst.SeekType.SET,
            int(self._state.position),
            Gst.SeekType.NONE,
            0,
        )
        return True

    def _is_eos(self) -> bool:
        assert self._state is not None
        return abs(self._state.position - self._state.duration) < 0.1e9

    def _query_position(self) -> float:
        assert self._playbin is not None
        assert self._state is not None

        if self._playbin.query(self._query) and (
            self._is_paused() or self._is_playing()
        ):
            _fmt, position = self._query.parse_position()
            return position

        return self._state.position

    def _map_gst_state_to_audio_state(self, gst_state: Gst.State) -> AudioPlayerState:
        if gst_state == Gst.State.NULL:
            return AudioPlayerState.NULL
        elif gst_state == Gst.State.READY:
            return AudioPlayerState.READY
        elif gst_state == Gst.State.PAUSED:
            return AudioPlayerState.PAUSED
        else:
            return AudioPlayerState.PLAYING

    def _setup_pipeline(self) -> None:
        assert self._playbin is not None
        # Set up the whole pipeline
        # For reference see
        # https://gstreamer.freedesktop.org/
        # documentation/audiofx/scaletempo.html
        audio_sink = Gst.Bin.new("audiosink")
        audioconvert = Gst.ElementFactory.make("audioconvert", "audioconvert")
        scaletempo = Gst.ElementFactory.make("scaletempo", "scaletempo")
        audioresample = Gst.ElementFactory.make("audioresample", "audioresample")
        autoaudiosink = Gst.ElementFactory.make("autoaudiosink", "autoaudiosink")

        required_elements = [audio_sink, audioconvert, audioresample, autoaudiosink]

        if any(element is None for element in required_elements):
            log.error("Could not set up full audio preview pipeline.")
        else:
            assert autoaudiosink is not None
            assert audioconvert is not None
            assert audioresample is not None

            # On Windows the first fraction of an audio
            # would not play if not synced.
            # On Linux there can be a delay before playback starts with sync,
            # which is however not the case on Windows.
            if sys.platform == "win32":
                autoaudiosink.set_property("sync", True)
            else:
                autoaudiosink.set_property("sync", False)

            audio_sink.add(audioconvert)
            audio_sink.add(audioresample)
            audio_sink.add(autoaudiosink)

            if scaletempo is not None:
                audio_sink.add(scaletempo)
                audioconvert.link(scaletempo)
                scaletempo.link(audioresample)
            else:
                # Without scaletempo, speed changes will cause a chipmunk effect
                log.warning("scaletempo not available")
                audioconvert.link(audioresample)

            audioresample.link(autoaudiosink)

            sink_pad = audioconvert.get_static_pad("sink")
            assert sink_pad is not None
            ghost_pad = Gst.GhostPad.new("sink", sink_pad)
            assert ghost_pad is not None
            audio_sink.add_pad(ghost_pad)
            self._playbin.set_property("audio-sink", audio_sink)

        self._playbin.no_more_pads()

        state_return = self._playbin.set_state(Gst.State.READY)
        if state_return == Gst.StateChangeReturn.FAILURE:
            log.error("Could not setup GST playbin.")
            return

        bus = self._playbin.get_bus()
        if bus is None:
            log.error("Could not get GST Bus.")
            return

        bus.add_signal_watch()
        self._bus_watch_id = bus.connect("message", self._on_bus_message)
        self._pipeline_is_setup = True

    def _is_paused(self) -> bool:
        assert self._playbin is not None
        _, state, _ = self._playbin.get_state(timeout=40)
        return state == Gst.State.PAUSED

    def _is_playing(self) -> bool:
        assert self._playbin is not None
        _, state, _ = self._playbin.get_state(timeout=40)
        return state == Gst.State.PLAYING

    def _clip_position(self, pos: float) -> float:
        assert self._state is not None
        if pos >= self._state.duration:
            return self._state.duration
        if pos < 0:
            return 0.0
        return pos

    def _set_playback_position(self, position: float) -> None:
        assert self._state is not None
        assert self._playbin is not None

        self._state.position = self._clip_position(position)
        self._query.set_position(Gst.Format.TIME, int(self._state.position))
        self._playbin.seek(
            self._state.speed,
            Gst.Format.TIME,
            Gst.SeekFlags.FLUSH,
            Gst.SeekType.SET,
            int(self._state.position),
            Gst.SeekType.NONE,
            0,
        )

    def _broadcast_playback_progression(self):
        if not self._do_broadcasting:
            return False

        assert self._state is not None
        self._state.position = self._query_position()
        self.emit(
            "audio-playback-progressed",
            self._preview_id,
            self._state.pipeline_state.value,
        )
        return True

    def _on_bus_message(self, _bus: Gst.Bus, message: Gst.Message) -> None:
        assert self._playbin is not None

        if (
            message.type == Gst.MessageType.ASYNC_DONE
            and self._file_path is not None
            and self._state is not None
        ):
            if self._switch_track:
                self._playbin.set_state(Gst.State.READY)
                self._playbin.set_property("uri", self._file_path.as_uri())
                self._query.set_position(Gst.Format.TIME, int(self._state.position))
                self._switch_track = False
                self._playbin.set_state(Gst.State.PAUSED)
            elif self._seek_track_position:
                self._set_playback_position(self._state.position)
                self._seek_track_position = False
            elif self._play_new_track:
                self._play_new_track = False
                self._playbin.set_state(Gst.State.PLAYING)

        if message.type == Gst.MessageType.STATE_CHANGED:
            old_state, new_state, _pending_state = message.parse_state_changed()
            if message.src == self._playbin and old_state != new_state:
                if self._state is None:
                    return

                self._state.pipeline_state = self._map_gst_state_to_audio_state(
                    new_state
                )
                self.emit(
                    "audio-playback-changed",
                    self._preview_id,
                    self._state.pipeline_state.value,
                )
                log.debug(
                    "State changed: %s → %s",
                    Gst.Element.state_get_name(old_state),
                    Gst.Element.state_get_name(new_state),
                )

                if new_state == Gst.State.PLAYING:
                    self._do_broadcasting = True
                    GLib.timeout_add(20, self._broadcast_playback_progression)
                    self.emit(
                        "audio-playback-changed",
                        self._preview_id,
                        self._state.pipeline_state.value,
                    )
                    return
                self._do_broadcasting = False
            return

        if message.type == Gst.MessageType.EOS and self._state is not None:
            if self._loop:
                self._set_playback_position(0.0)
                self._playbin.set_state(Gst.State.PLAYING)
            else:
                self._state.position = self._state.duration
                self._playbin.set_state(Gst.State.PAUSED)
                self._do_broadcasting = False
            return

        if message.type == Gst.MessageType.ERROR:
            self._handle_error_on_playback(message)

    def _error_info(self, message: Gst.Message) -> tuple[str, int]:
        structure = message.get_structure()
        assert structure is not None
        gerror = structure.get_value("gerror")
        assert gerror is not None
        return gerror.domain, gerror.code

    def _handle_error_on_playback(self, message: Gst.Message):
        assert self._playbin is not None
        domain, code = self._error_info(message)
        if domain == "gst-resource-error-quark":
            if code == 10:
                # On Windows: AUDCLNT_E_DEVICE_INVALIDATED (10)
                # GST_RESOURCE_ERROR_WRITE (10)
                # used when the resource can't be written to.
                self._playbin.set_state(Gst.State.READY)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import GdkPixbuf
from gi.repository import Gtk

from gajim.common import app
from gajim.common import types
from gajim.common.const import AvatarSize
from gajim.common.events import JingleRequestReceived
from gajim.common.i18n import _
from gajim.common.jingle_session import JingleSession
from gajim.common.modules.contacts import BareContact
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.models import Message
from gajim.common.util.datetime import utc_now

from gajim.gtk.conversation.rows.base import BaseRow
from gajim.gtk.conversation.rows.widgets import DateTimeLabel
from gajim.gtk.conversation.rows.widgets import NicknameLabel
from gajim.gtk.conversation.rows.widgets import SimpleLabel


class CallRow(BaseRow):
    def __init__(
        self,
        account: str,
        contact: types.BareContact,
        event: JingleRequestReceived | None = None,
        message: Message | None = None,
    ) -> None:
        BaseRow.__init__(self, account)

        self.type = "call"

        self._client = app.get_client(account)

        if message is not None:
            timestamp = message.timestamp
        else:
            timestamp = utc_now()
        self.timestamp = timestamp.astimezone()
        self.db_timestamp = timestamp.timestamp()

        self._contact = contact
        self._event = event
        self._message = message

        self._session: JingleSession | None = None

        if message is not None and message.call is not None:
            module = self._client.get_module("Jingle")
            self._session = module.get_jingle_session(
                str(self._contact.jid), message.call.sid
            )
            self.pk = message.pk

        self._avatar_placeholder = Gtk.Box()
        self._avatar_placeholder.set_size_request(AvatarSize.ROSTER, -1)
        self.grid.attach(self._avatar_placeholder, 0, 0, 1, 1)

        if event is None and self._session is None:
            self._add_history_call_widget()
        else:
            self._add_incoming_call_widget()

    def do_unroot(self) -> None:
        BaseRow.do_unroot(self)

    def update(self) -> None:
        if self._event is None and self._session is None:
            return

        self.grid.remove(self._call_box)
        self._add_history_call_widget()

        self._event = None
        self._session = None

    def _on_accept(self, button: Gtk.Button) -> None:
        button.set_sensitive(False)
        self._decline_button.set_sensitive(False)
        if self._event is not None:
            session = self._client.get_module("Jingle").get_jingle_session(
                self._event.fjid, self._event.sid
            )
            if session is not None:
                app.call_manager.accept_call(session)
        else:
            assert self._session is not None
            app.call_manager.accept_call(self._session)

    def _on_decline(self, button: Gtk.Button) -> None:
        button.set_sensitive(False)
        self._accept_button.set_sensitive(False)
        if self._event is not None:
            session = self._client.get_module("Jingle").get_jingle_session(
                self._event.fjid, self._event.sid
            )
            if session is not None:
                app.call_manager.decline_call(session)
        else:
            assert self._session is not None
            app.call_manager.decline_call(self._session)
        self._session = None

    def _add_history_call_widget(self) -> None:
        contact = self._client.get_module("Contacts").get_contact(
            self._client.get_own_jid().bare
        )
        assert isinstance(contact, BareContact)

        is_self = True
        if self._message is not None:
            if self._message.direction == ChatDirection.INCOMING:
                contact = self._contact
                is_self = True
            else:
                is_self = False

        if self._event is not None:
            is_self = False
        else:
            contact = self._contact
            is_self = True

        scale = self.get_scale_factor()
        avatar = contact.get_avatar(AvatarSize.ROSTER, scale, add_show=False)
        assert not isinstance(avatar, GdkPixbuf.Pixbuf)
        avatar_image = Gtk.Image.new_from_paintable(avatar)
        self._avatar_placeholder.append(avatar_image)

        name_widget = NicknameLabel(contact.name, is_self)
        name_widget.set_halign(Gtk.Align.START)
        name_widget.set_valign(Gtk.Align.START)

        timestamp_widget = DateTimeLabel(self.timestamp)
        timestamp_widget.set_halign(Gtk.Align.START)
        timestamp_widget.set_valign(Gtk.Align.START)

        meta_box = Gtk.Box()
        meta_box.set_spacing(6)
        meta_box.append(name_widget)
        meta_box.append(timestamp_widget)
        self.grid.attach(meta_box, 1, 0, 1, 1)

        icon = Gtk.Image.new_from_icon_name("lucide-phone-symbolic")

        label = SimpleLabel()
        label.add_css_class("dimmed")
        label.set_text(_("Call"))

        content_box = Gtk.Box(spacing=12)
        content_box.append(icon)
        content_box.append(label)
        self.grid.attach(content_box, 1, 1, 1, 1)

    def _add_incoming_call_widget(self) -> None:
        self._call_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)
        self._call_box.set_size_request(350, -1)
        self._call_box.add_css_class("card")
        self._call_box.add_css_class("p-12")
        self._call_box.add_css_class("conversation-call-box")

        scale = self.get_scale_factor()
        avatar = self._contact.get_avatar(AvatarSize.CALL, scale, add_show=False)
        assert not isinstance(avatar, GdkPixbuf.Pixbuf)
        avatar_image = Gtk.Image.new_from_paintable(avatar)
        self._call_box.append(avatar_image)

        content_types: list[str] = []
        if self._event is not None:
            for item in self._event.contents:
                content_types.append(item.media)
        if self._session is not None:
            if self._session.get_content("audio") is not None:
                content_types.append("audio")
            if self._session.get_content("video") is not None:
                content_types.append("video")

        text = _("%s is calling") % self._contact.name
        if "video" in content_types:
            text += _(" (Video Call)")
        else:
            text += _(" (Voice Call)")
        label = Gtk.Label(
            label=text,
            max_width_chars=40,
            wrap=True,
        )
        label.add_css_class("bold")

        self._call_box.append(label)

        self._decline_button = Gtk.Button()
        if self._session is not None:
            self._decline_button.set_sensitive(not self._session.accepted)
        self._decline_button.add_css_class("destructive-action")
        self._connect(self._decline_button, "clicked", self._on_decline)
        decline_icon = Gtk.Image.new_from_icon_name("lucide-phone-off-symbolic")
        self._decline_button.set_child(decline_icon)
        decline_label = Gtk.Label(label=_("Decline"))
        decline_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=3)
        decline_box.append(self._decline_button)
        decline_box.append(decline_label)

        self._accept_button = Gtk.Button()
        if self._session is not None:
            self._accept_button.set_sensitive(not self._session.accepted)
        self._accept_button.add_css_class("suggested-action")
        self._connect(self._accept_button, "clicked", self._on_accept)
        accept_icon = Gtk.Image.new_from_icon_name("lucide-phone-symbolic")
        self._accept_button.set_child(accept_icon)
        accept_label = Gtk.Label(label=_("Accept"))
        accept_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=3)
        accept_box.append(self._accept_button)
        accept_box.append(accept_label)

        button_box = Gtk.Box(spacing=50)
        button_box.set_halign(Gtk.Align.CENTER)
        button_box.append(decline_box)
        button_box.append(accept_box)
        self._call_box.append(button_box)

        self.grid.attach(self._call_box, 1, 0, 1, 1)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import logging
import math
from dataclasses import dataclass

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import Graphene
from gi.repository import Gsk
from gi.repository import Gtk

from gajim.common import app

from gajim.gtk.audio_player import AudioSampleT

log = logging.getLogger("gajim.gtk.preview_audio_visualizer")


@dataclass(frozen=True)
class SliderStyle:
    fill_color: Gdk.RGBA
    border_color: Gdk.RGBA
    border_width: float


@dataclass(frozen=True)
class SliderGeometry:
    x: float
    y: float
    radius: float


class AudioVisualizerWidget(Gtk.Widget):
    _MIN_WIDTH = 80

    def __init__(
        self,
        width: int = 340,
        height: int = 50,
        x_offset: int = 0,
        is_seekable: bool = False,
    ) -> None:
        Gtk.Widget.__init__(self, margin_start=x_offset)

        self._bar_width = 2  # px
        self._natural_width = width
        self._natural_height = height

        self._width = width
        self._height = height
        self._is_LTR = bool(self.get_direction() == Gtk.TextDirection.LTR)

        self._raw_samples: AudioSampleT = []
        self._samples: AudioSampleT = []
        self._is_seekable = is_seekable
        self._seek_position = -1.0
        self._position = 0.0
        self._live_mode = False
        self._scroll_fraction = 0.0

        self._waveform_path: Gsk.Path | None = None

        accent = Adw.StyleManager.get_default().get_accent_color_rgba()

        style_context = self.get_style_context()
        self._slider_color = style_context.lookup_color("window_bg_color")[1]
        self._slider_border_color = style_context.lookup_color("window_fg_color")[1]

        self._color_progress = accent
        self._color_seek = Gdk.RGBA(
            red=min(1.0, accent.red * 1.3),
            green=min(1.0, accent.green * 1.3),
            blue=min(1.0, accent.blue * 1.3),
            alpha=accent.alpha,
        )
        color_default = accent.copy()
        color_default.alpha = max(0.0, accent.alpha - 0.4)
        self._color_default = color_default

    def run_destroy(self) -> None:
        app.check_finalize(self)

    def do_measure(
        self, orientation: Gtk.Orientation, for_size: int
    ) -> tuple[int, int, int, int]:
        if orientation == Gtk.Orientation.HORIZONTAL:
            return self._MIN_WIDTH, self._natural_width, -1, -1

        return self._natural_height, self._natural_height, -1, -1

    def do_size_allocate(self, width: int, height: int, baseline: int) -> None:
        if width == self._width and height == self._height:
            return

        self._width = width
        self._height = height
        self._rebuild_static_path()

    def get_effective_width(self) -> int:
        return self._width

    @property
    def _slider_radius(self) -> int:
        return self._height // 5

    @property
    def _graph_x_offset(self) -> float:
        # Inset the graph by the slider's radius, so that the slider can be
        # centered on the first and the last bar without leaving our bounds.
        if not self._is_seekable:
            return 0.0
        return float(self._slider_radius)

    @property
    def _graph_width(self) -> float:
        return self._width - 2 * self._graph_x_offset

    def convert_x_to_position(self, x: float) -> float:
        if self._graph_width <= 0:
            return 0.0

        position = (x - self._graph_x_offset) / self._graph_width
        return min(max(position, 0.0), 1.0)

    def set_parameters(self, position: float, live_mode: bool = False) -> None:
        self._position = position
        self._live_mode = live_mode

    def set_samples(self, samples: AudioSampleT) -> None:
        self._raw_samples = samples
        self._samples = samples
        self._rebuild_static_path()

    def _rebuild_static_path(self) -> None:
        if not self._is_static or not self._raw_samples:
            return

        samples = self._downsample(self._raw_samples)
        samples = self._normalize(samples)
        samples = self._rescale(samples)
        self._samples = samples
        self._waveform_path = self._build_static_path(samples)

    def render_animated_graph(self, scroll_fraction: float = 0.0) -> None:
        if not self._samples:
            log.debug("Render animated graph: No samples")
            return
        self._scroll_fraction = scroll_fraction
        self.queue_draw()

    def render_static_graph(self, position: float, seek_position: float = -1.0) -> None:
        if not self._samples:
            return
        self._position = position
        self._seek_position = seek_position
        self.queue_draw()

    def do_snapshot(self, snapshot: Gtk.Snapshot) -> None:
        if self._is_static:
            if self._waveform_path is None:
                log.debug("Waveform path is None")
                return
            self._draw_static(snapshot)
        else:
            if not self._samples:
                return
            self._draw_animated(snapshot)

    def _draw_static(self, snapshot: Gtk.Snapshot) -> None:
        assert self._waveform_path is not None
        if not self._is_LTR:
            self._apply_rtl_transform(snapshot)

        play_x = self._graph_x_offset + self._position * self._graph_width

        snapshot.push_clip(Graphene.Rect().init(0, 0, self._width, self._height))
        snapshot.append_fill(
            self._waveform_path, Gsk.FillRule.WINDING, self._color_default
        )
        snapshot.pop()

        snapshot.push_clip(Graphene.Rect().init(0, 0, play_x, self._height))
        snapshot.append_fill(
            self._waveform_path, Gsk.FillRule.WINDING, self._color_progress
        )
        snapshot.pop()

        if self._seek_position >= 0:
            seek_x = self._graph_x_offset + self._seek_position * self._graph_width
            x0 = min(play_x, seek_x)
            x1 = max(play_x, seek_x)
            snapshot.push_clip(Graphene.Rect().init(x0, 0, x1 - x0, self._height))
            snapshot.append_fill(
                self._waveform_path, Gsk.FillRule.WINDING, self._color_seek
            )
            snapshot.pop()
            slider_pos = seek_x
        else:
            slider_pos = play_x

        # The position may run out of bounds while seeking
        slider_pos = max(slider_pos, self._graph_x_offset)
        slider_pos = min(slider_pos, self._graph_x_offset + self._graph_width)
        self._draw_slider(
            snapshot,
            SliderGeometry(
                x=slider_pos, y=self._height // 2, radius=self._slider_radius
            ),
            SliderStyle(
                fill_color=self._slider_color,
                border_color=self._slider_border_color,
                border_width=1,
            ),
        )

    def _draw_animated(self, snapshot: Gtk.Snapshot) -> None:
        if not self._is_LTR:
            self._apply_rtl_transform(snapshot)

        n = len(self._samples)
        x_step = (self._graph_width - 2 * self._bar_width) / n
        x0 = self._graph_x_offset + x_step * (1.0 - self._scroll_fraction)
        cy = self._height / 2
        fade_n = max(1, n // 8)

        # Batch all fully opaque bars into a single path and draw call
        pb = Gsk.PathBuilder.new()
        for i in range(fade_n, n):
            self._add_bar(pb, x0 + i * x_step, cy, *self._samples[i])
        snapshot.append_fill(pb.to_path(), Gsk.FillRule.WINDING, self._color_progress)

        # Draw faded bars individually — only ~10 of them
        r = self._color_progress.red
        g = self._color_progress.green
        b = self._color_progress.blue
        a = self._color_progress.alpha
        for i in range(fade_n):
            t = i / fade_n
            alpha = t * t * (3.0 - 2.0 * t)  # smoothstep fade-in
            color = Gdk.RGBA(red=r, green=g, blue=b, alpha=a * alpha)
            pb = Gsk.PathBuilder.new()
            self._add_bar(pb, x0 + i * x_step, cy, *self._samples[i])
            snapshot.append_fill(pb.to_path(), Gsk.FillRule.WINDING, color)

    def _apply_rtl_transform(self, snapshot: Gtk.Snapshot) -> None:
        cx, cy = self._width / 2, self._height / 2
        snapshot.translate(Graphene.Point().init(cx, cy))
        snapshot.rotate(180)
        snapshot.translate(Graphene.Point().init(-cx, -cy))

    def _draw_slider(
        self,
        snapshot: Gtk.Snapshot,
        geometry: SliderGeometry,
        style: SliderStyle,
    ) -> None:
        x = geometry.x
        y = geometry.y
        radius = geometry.radius
        fill_color = style.fill_color
        border_color = style.border_color
        border_width = style.border_width

        rect = Graphene.Rect()
        rect.init(
            x - radius,
            y - radius,
            radius * 2,
            radius * 2,
        )
        rounded = Gsk.RoundedRect()
        rounded.init_from_rect(rect, radius)
        snapshot.push_rounded_clip(rounded)
        snapshot.append_color(fill_color, rect)
        snapshot.pop()
        snapshot.append_border(
            rounded,
            [border_width] * 4,
            [border_color] * 4,
        )

    def _build_static_path(self, samples: AudioSampleT) -> Gsk.Path | None:
        if not samples:
            return None
        n = len(samples)
        x_step = (self._graph_width - 2 * self._bar_width) / n
        cy = self._height / 2
        pb = Gsk.PathBuilder.new()
        for i, (s1, s2) in enumerate(samples):
            self._add_bar(pb, self._graph_x_offset + x_step * (i + 1), cy, s1, s2)
        return pb.to_path()

    def _add_bar(
        self,
        pb: Gsk.PathBuilder,
        x: float,
        cy: float,
        h1: float,
        h2: float,
    ) -> None:
        # Upper half grows downward from cy, lower half grows upward.
        # Outer (top/bottom) edges are rounded with a conic arc.
        scaling = self._height / 2 - 1
        h1 = max(0.5, h1 * scaling)
        h2 = max(0.5, h2 * scaling)
        hw = self._bar_width / 2
        r = self._bar_width  # right edge x

        pb.move_to(x, cy + h1)
        pb.conic_to(x + hw, cy + h1 + hw, x + r, cy + h1, 30)
        pb.line_to(x + r, cy)
        pb.line_to(x, cy)
        pb.close()

        pb.move_to(x + r, cy)
        pb.line_to(x + r, cy - h2)
        pb.conic_to(x + hw, cy - h2 - hw, x, cy - h2, 30)
        pb.line_to(x, cy)
        pb.line_to(x + r, cy)
        pb.close()

    def _downsample(self, samples: AudioSampleT) -> AudioSampleT:
        num_bars = int(self._graph_width / (self._bar_width * 2))
        if num_bars == 0:
            log.error("Number of bars is zero")
            return []

        stride = math.floor(len(samples) / num_bars) + 1
        if stride < 2:
            return samples

        s1_vals, s2_vals = zip(*samples, strict=True)
        result: AudioSampleT = []
        for i in range(2, int(len(samples) - stride / 2), stride):
            c = int(i + stride / 2)
            result.append(
                (
                    (s1_vals[c - 1] + s1_vals[c]) / 2,
                    (s2_vals[c - 1] + s2_vals[c]) / 2,
                )
            )
        return result

    def _normalize(self, samples: AudioSampleT) -> AudioSampleT:
        if not samples:
            log.error("No samples to normalize")
            return []
        all_vals = [v for pair in samples for v in pair]
        lo = min(all_vals)
        hi = max(all_vals)
        span = hi - lo
        if span <= 0:
            return samples
        return [((v1 - lo) / span, (v2 - lo) / span) for v1, v2 in samples]

    def _rescale(self, samples: AudioSampleT) -> AudioSampleT:
        # sin(sqrt) expands quieter peaks to match perceptual loudness
        inv_sin1 = 1.0 / math.sin(1)
        return [
            (
                math.sin(math.sqrt(abs(v1))) * inv_sin1,
                math.sin(math.sqrt(abs(v2))) * inv_sin1,
            )
            for v1, v2 in samples
        ]

    @property
    def _is_static(self) -> bool:
        return not self._live_mode



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0060: Publish-Subscribe

from __future__ import annotations

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import DiscoInfo

from gajim.common import types
from gajim.common.modules.base import BaseModule


class PubSub(BaseModule):
    _nbxmpp_extends = "PubSub"
    _nbxmpp_methods = [
        "publish",
        "delete",
        "set_node_configuration",
        "get_node_configuration",
        "get_access_model",
        "request_items",
        "subscribe",
        "unsubscribe",
        "get_subscriptions",
    ]

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.publish_options = False

    def pass_disco(self, info: DiscoInfo) -> None:
        if Namespace.PUBSUB_PUBLISH_OPTIONS in info.features:
            self._log.info("Discovered Pubsub publish options: %s", info.jid)
            self.publish_options = True



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import sys

from gi.repository import Adw
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from nbxmpp.protocol import JID

from gajim.common import app
from gajim.common.client import Client
from gajim.common.const import AvatarSize
from gajim.common.const import SimpleClientState
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import ResourceContact

from gajim.gtk import structs
from gajim.gtk.activity_list import ActivityListItem
from gajim.gtk.activity_list import ActivityListItemT
from gajim.gtk.activity_list import GajimPluginUpdate
from gajim.gtk.activity_list import GajimPluginUpdateFinished
from gajim.gtk.activity_list import GajimUpdate
from gajim.gtk.activity_list import GajimUpdatePermission
from gajim.gtk.activity_list import MucInvitation
from gajim.gtk.activity_list import MucInvitationDeclined
from gajim.gtk.activity_list import OpenPGPEvent
from gajim.gtk.activity_list import Reaction
from gajim.gtk.activity_list import Subscribe
from gajim.gtk.activity_list import TimezoneChanged
from gajim.gtk.activity_list import Unsubscribed
from gajim.gtk.builder import get_builder
from gajim.gtk.groupchat_invitation import GroupChatInvitation
from gajim.gtk.menus import get_subscription_menu
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import check_finalize
from gajim.gtk.util.misc import open_uri
from gajim.gtk.util.window import open_window


class ActivityPage(Gtk.Stack):
    __gsignals__ = {
        "page-removed": (GObject.SignalFlags.RUN_LAST, None, (ActivityListItem,)),
    }

    def __init__(self) -> None:
        Gtk.Stack.__init__(self, hexpand=True)
        self.add_css_class("activity-page")
        self.add_named(DefaultPage(), "default")

        self._pages = {
            GajimUpdate: GajimUpdatePage,
            GajimPluginUpdate: GajimPluginUpdatePage,
            GajimUpdatePermission: GajimUpdatePermissionPage,
            GajimPluginUpdateFinished: GajimPluginUpdateFinishedPage,
            Subscribe: SubscribePage,
            Unsubscribed: UnsubscribedPage,
            MucInvitation: InvitationPage,
            MucInvitationDeclined: InvitationDeclinedPage,
            OpenPGPEvent: OpenPGPEventPage,
            TimezoneChanged: TimezoneChangedPage,
        }

    def show_default_page(self) -> None:
        self.set_visible_child_name("default")

    def process_row_activated(self, item: ActivityListItemT) -> None:
        child = self.get_child_by_name("activity")
        if child is not None:
            self.remove(child)

        if isinstance(item, Reaction):
            return

        page = self._pages[type(item)](item)  # type: ignore
        page.connect("request-remove", self._on_request_remove)

        self.add_named(page, "activity")
        self.set_visible_child_name("activity")

    def _on_request_remove(self, page: BaseActivityPage) -> None:
        self.show_default_page()
        item = page.get_item()
        self.remove(page)
        self.emit("page-removed", item)


class BaseActivityPage(Gtk.Box, SignalManager):
    __gsignals__ = {
        "request-remove": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, item: ActivityListItemT | None = None) -> None:
        Gtk.Box.__init__(self)
        SignalManager.__init__(self)
        self.add_css_class("mt-18")

        self._ui = None
        self._item = item

        clamp = Adw.Clamp(hexpand=True)
        self.append(clamp)

        self._content_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)
        clamp.set_child(self._content_box)

    def add_widget(self, widget: Gtk.Widget) -> None:
        self._content_box.append(widget)

    def get_item(self) -> ActivityListItemT:
        assert self._item is not None
        return self._item

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.Box.do_unroot(self)
        del self._ui
        check_finalize(self)


class DefaultPage(BaseActivityPage):
    def __init__(self) -> None:
        BaseActivityPage.__init__(self)
        self._ui = get_builder("activity_default.ui")
        self.add_widget(self._ui.default_page)


class GajimUpdatePermissionPage(BaseActivityPage):
    def __init__(self, item: GajimUpdatePermission) -> None:
        BaseActivityPage.__init__(self)
        self._ui = get_builder("activity_gajim_update.ui")

        self._ui.update_permission_box.set_visible(True)
        self._ui.update_permission_title_label.set_text(item.title)
        self._ui.update_permission_enable_button.set_sensitive(
            item.state.get("state") != "completed"
        )
        self._connect(
            self._ui.update_permission_enable_button,
            "clicked",
            self._on_enable_update_clicked,
            item,
        )
        self.add_widget(self._ui.gajim_update_page)

    def _on_enable_update_clicked(
        self, _button: Gtk.Button, item: GajimUpdatePermission
    ) -> None:
        app.settings.set("check_for_update", True)
        app.app.check_for_gajim_updates()
        item.state["state"] = "completed"
        self._ui.update_permission_enable_button.set_sensitive(False)


class GajimUpdatePage(BaseActivityPage):
    def __init__(self, item: GajimUpdate) -> None:
        BaseActivityPage.__init__(self)
        self._ui = get_builder("activity_gajim_update.ui")

        self._ui.update_gajim_box.set_visible(True)
        self._ui.update_gajim_title_label.set_text(item.title)
        self._ui.update_gajim_text_label.set_text(item.subject)
        self._connect(
            self._ui.update_gajim_button,
            "clicked",
            self._on_download_update_clicked,
            item,
        )
        self.add_widget(self._ui.gajim_update_page)

    def _on_download_update_clicked(
        self, _button: Gtk.Button, item: GajimUpdate
    ) -> None:
        if sys.platform == "win32":
            event = item.get_event()
            open_uri(event.setup_url)
        else:
            open_uri("https://gajim.org/download/")


class GajimPluginUpdatePage(BaseActivityPage):
    def __init__(self, item: GajimPluginUpdate) -> None:
        BaseActivityPage.__init__(self)
        self._ui = get_builder("activity_gajim_update.ui")

        self._ui.update_plugins_box.set_visible(True)
        self._ui.update_plugins_title_label.set_text(item.title)
        self._ui.update_plugins_text_label.set_text(item.subject)
        self._connect(
            self._ui.update_plugins_show_button,
            "clicked",
            self._on_open_plugins,
            item,
        )

        self._ui.update_plugins_button.set_sensitive(
            item.state.get("state") != "completed"
        )
        self._connect(
            self._ui.update_plugins_button,
            "clicked",
            self._on_update_plugins_clicked,
            item,
        )
        self.add_widget(self._ui.gajim_update_page)

    def _on_update_plugins_clicked(
        self, _button: Gtk.Button, item: GajimPluginUpdate
    ) -> None:
        app.settings.set(
            "plugins_auto_update",
            self._ui.update_plugins_automatically_checkbox.get_active(),
        )

        event = item.get_event()
        app.plugin_repository.download_plugins(event.manifests)

        item.state["state"] = "completed"
        self._ui.update_plugins_button.set_sensitive(False)

    def _on_open_plugins(self, _button: Gtk.Button, item: GajimPluginUpdate) -> None:
        app.app.activate_action("plugins", None)


class GajimPluginUpdateFinishedPage(BaseActivityPage):
    def __init__(self, item: GajimPluginUpdateFinished) -> None:
        BaseActivityPage.__init__(self)
        self._ui = get_builder("activity_gajim_update.ui")

        self._ui.update_plugins_success_box.set_visible(True)
        self._ui.update_plugins_success_title_label.set_text(item.title)
        self._ui.update_plugins_success_text_label.set_text(item.subject)
        self._connect(
            self._ui.update_plugins_success_button,
            "clicked",
            self._on_open_plugins,
            item,
        )
        self.add_widget(self._ui.gajim_update_page)

    def _on_open_plugins(
        self, _button: Gtk.Button, item: GajimPluginUpdateFinished
    ) -> None:
        app.app.activate_action("plugins", None)


class SubscribePage(BaseActivityPage):
    def __init__(self, item: Subscribe) -> None:
        BaseActivityPage.__init__(self)
        self._ui = get_builder("activity_subscription.ui")

        self._ui.subscription_title_label.set_text(item.title)
        self._ui.subscription_text_label.set_text(item.subject)

        event = item.get_event()

        client = app.get_client(event.account)
        contact = client.get_module("Contacts").get_contact(event.jid)
        assert isinstance(contact, BareContact)

        self._ui.subscription_image.set_from_paintable(
            contact.get_avatar(
                AvatarSize.ACTIVITY_PAGE, self.get_scale_factor(), add_show=False
            )
        )

        self._ui.subscribe_box.set_sensitive(item.state.get("state") != "completed")
        self._ui.subscribe_box.set_visible(True)

        jid = JID.from_string(event.jid)

        deny_param = structs.AccountJidParam(account=event.account, jid=jid)
        self._ui.subscribe_deny_button.set_action_target_value(deny_param.to_variant())
        self._ui.subscribe_deny_button.set_action_name(
            f"app.{event.account}-subscription-deny"
        )
        self._connect(self._ui.subscribe_deny_button, "clicked", self._on_click, item)

        accept_param = structs.SubscriptionAcceptParam(
            account=event.account, jid=jid, nickname=event.user_nick
        )
        self._ui.subscribe_accept_button.set_action_target_value(
            accept_param.to_variant()
        )
        self._ui.subscribe_accept_button.set_action_name(
            f"app.{event.account}-subscription-accept"
        )
        self._connect(self._ui.subscribe_deny_button, "clicked", self._on_click, item)

        self._ui.subscribe_menu_button.set_menu_model(
            get_subscription_menu(event.account, jid)
        )

        self.add_widget(self._ui.subscription_page)

    def _on_click(self, button: Gtk.Button, item: Subscribe) -> None:
        item.state["state"] = "completed"
        self._ui.subscribe_box.set_sensitive(False)


class UnsubscribedPage(BaseActivityPage):
    def __init__(self, item: Unsubscribed) -> None:
        BaseActivityPage.__init__(self)
        self._ui = get_builder("activity_subscription.ui")

        self._ui.subscription_title_label.set_text(item.title)
        self._ui.subscription_text_label.set_text(item.subject)

        event = item.get_event()

        client = app.get_client(event.account)
        contact = client.get_module("Contacts").get_contact(event.jid)
        assert isinstance(contact, BareContact)

        self._ui.subscription_image.set_from_paintable(
            contact.get_avatar(
                AvatarSize.ACTIVITY_PAGE, self.get_scale_factor(), add_show=False
            )
        )

        self._ui.unsubscribed_box.set_sensitive(item.state.get("state") != "completed")
        self._ui.unsubscribed_box.set_visible(True)
        self._ui.unsubscribed_remove_button.set_action_target_value(
            GLib.Variant("as", [event.account, str(event.jid)])
        )
        self._ui.unsubscribed_remove_button.set_action_name(
            f"app.{event.account}-remove-contact"
        )
        self._connect(
            self._ui.unsubscribed_remove_button, "clicked", self._on_clicked, item
        )

        self.add_widget(self._ui.subscription_page)

    def _on_clicked(self, _button: Gtk.Button, item: Unsubscribed) -> None:
        item.state["state"] = "completed"
        self._ui.unsubscribed_box.set_sensitive(False)


class InvitationPage(BaseActivityPage):
    def __init__(self, item: MucInvitation) -> None:
        BaseActivityPage.__init__(self)
        self._ui = get_builder("activity_muc_invitation.ui")

        self._ui.muc_invitation_box.set_visible(True)
        event = item.get_event()
        invitation_widget = GroupChatInvitation(event.account, event)

        if item.state.get("state") == "completed":
            invitation_widget.disable_buttons()
        else:
            self._connect(invitation_widget, "declined", self._on_clicked, item)
            self._connect(invitation_widget, "accepted", self._on_clicked, item)

        self._ui.muc_invitation_box.append(invitation_widget)
        self.add_widget(self._ui.muc_invitation_page)

    def _on_clicked(
        self, invitation_widget: GroupChatInvitation, item: MucInvitation
    ) -> None:
        item.state["state"] = "completed"
        invitation_widget.disable_buttons()


class InvitationDeclinedPage(BaseActivityPage):
    def __init__(self, item: MucInvitationDeclined) -> None:
        BaseActivityPage.__init__(self)
        self._ui = get_builder("activity_muc_invitation.ui")

        self._ui.muc_invitation_declined_box.set_visible(True)
        self._ui.invitation_title_label.set_text(item.title)
        self._ui.invitation_text_label.set_text(item.subject)

        event = item.get_event()
        client = app.get_client(event.account)
        contact = client.get_module("Contacts").get_contact(event.from_.bare)
        assert not isinstance(contact, ResourceContact)

        texture = contact.get_avatar(AvatarSize.ACCOUNT_PAGE, self.get_scale_factor())
        self._ui.invitation_image.set_from_paintable(texture)
        self.add_widget(self._ui.muc_invitation_page)


class OpenPGPEventPage(BaseActivityPage):
    def __init__(self, item: OpenPGPEvent) -> None:
        BaseActivityPage.__init__(self, item)
        self._ui = get_builder("activity_openpgp_event.ui")

        self._connect(self._ui.disable_button, "clicked", self._on_disable)
        self._connect(
            self._ui.backup_button,
            "clicked",
            self._on_setup_backup_clicked,
            "test-password",
        )
        self._connect(
            self._ui.setup_button, "clicked", self._on_setup_backup_clicked, None
        )

        self._event = item.get_event()
        self._client = app.get_client(self._event.account)
        self._client.connect_signal("state-changed", self._on_client_state_changed)

        if self._event.type == "setup":
            self._ui.title.set_text(_("OpenPGP Setup"))
            self._ui.text.set_text(_("Complete the OpenPGP setup to chat encrypted."))
            self._ui.setup_button.set_visible(True)
        else:
            self._ui.title.set_text(_("OpenPGP Key Backup"))
            self._ui.text.set_text(
                _(
                    "To secure your OpenPGP key for encrypted chats, Gajim can create an encrypted backup of your key on your server."
                )
            )
            self._ui.backup_button.set_visible(True)

        self.add_widget(self._ui.text_box)
        self._set_sensitive(self._client.is_available())

    def _set_sensitive(self, sensitive: bool) -> None:
        self._ui.backup_button.set_sensitive(sensitive)
        self._ui.disable_button.set_sensitive(sensitive)
        self._ui.setup_button.set_sensitive(sensitive)

    def _on_setup_backup_clicked(self, _button: Gtk.Button, mode: str | None) -> None:
        open_window("OpenPGPWizard", account=self._event.account, mode=mode)
        self.emit("request-remove")

    def _on_disable(self, _button: Gtk.Button) -> None:
        app.settings.set_account_setting(
            self._event.account, "openpgp_backup_secret_key", False
        )
        self.emit("request-remove")

    def _on_client_state_changed(
        self, client: Client, _signal_name: str, state: SimpleClientState
    ) -> None:
        self._set_sensitive(state.is_connected)


class TimezoneChangedPage(BaseActivityPage):
    def __init__(self, item: TimezoneChanged) -> None:
        BaseActivityPage.__init__(self, item)
        self._ui = get_builder("activity_change_timezone.ui")

        self._connect(self._ui.update_button, "clicked", self._on_update)
        self._connect(self._ui.disable_button, "clicked", self._on_disable)

        self._event = item.get_event()
        self._client = app.get_client(self._event.account)
        self._client.connect_signal("state-changed", self._on_client_state_changed)

        self._ui.old_timezone_label.set_text(self._event.vcard or _("No Timezone"))
        self._ui.new_timezone_label.set_text(self._event.local or "")

        self.add_widget(self._ui.change_timezone_page)
        self._set_sensitive(self._client.is_available())

    def _set_sensitive(self, sensitive: bool) -> None:
        self._ui.update_button.set_sensitive(sensitive)
        self._ui.disable_button.set_sensitive(sensitive)

    def _on_update(self, _button: Gtk.Button) -> None:
        self._client.get_module("VCard4").update_timezone()
        self.emit("request-remove")

    def _on_disable(self, _button: Gtk.Button) -> None:
        app.settings.set_account_setting(self._event.account, "update_timezone", False)
        self.emit("request-remove")

    def _on_client_state_changed(
        self, client: Client, _signal_name: str, state: SimpleClientState
    ) -> None:
        self._set_sensitive(state.is_connected)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from gi.repository import Gtk
from nbxmpp.protocol import validate_resourcepart

from gajim.common import app
from gajim.common.i18n import _

from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_ui_string


@Gtk.Template(string=get_ui_string("groupchat_nick_chooser.ui"))
class GroupChatNickChooser(Gtk.MenuButton, SignalManager):
    __gtype_name__ = "GroupChatNickChooser"

    _label: Gtk.Label = Gtk.Template.Child()
    _popover: Gtk.Popover = Gtk.Template.Child()
    _entry: Gtk.Entry = Gtk.Template.Child()
    _apply_button: Gtk.Button = Gtk.Template.Child()

    def __init__(self) -> None:
        Gtk.MenuButton.__init__(self)
        SignalManager.__init__(self)

        self._connect(self, "notify::active", self._on_nickname_button_toggled)
        self._connect(self._entry, "changed", self._on_nickname_changed)
        self._connect(self._apply_button, "clicked", self._on_apply_nickname)

    def do_unroot(self) -> None:
        Gtk.MenuButton.do_unroot(self)
        self._disconnect_all()
        app.check_finalize(self)

    def get_text(self) -> str:
        return self._entry.get_text()

    def set_text(self, text: str) -> None:
        self._entry.set_text(text)
        self._label.set_text(text)

    def _on_nickname_button_toggled(
        self, _widget: Gtk.MenuButton, _active: bool
    ) -> None:
        self._entry.grab_focus()

    def _on_nickname_changed(self, entry: Gtk.Entry) -> None:
        try:
            validate_resourcepart(entry.get_text())
            self._apply_button.set_sensitive(True)
            self._apply_button.set_tooltip_text(None)
        except Exception:
            self._apply_button.set_sensitive(False)
            self._apply_button.set_tooltip_text(
                _("Nickname contains invalid characters")
            )

    def _on_apply_nickname(self, _button: Gtk.Button) -> None:
        nickname = self._entry.get_text()
        self._popover.popdown()
        self._label.set_text(nickname)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0084: User Avatar

from __future__ import annotations

from typing import Any

from collections.abc import Generator

from nbxmpp.errors import is_error
from nbxmpp.modules.user_avatar import AvatarData
from nbxmpp.modules.user_avatar import AvatarMetaData
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.structs import MessageProperties

from gajim.common import app
from gajim.common import types
from gajim.common.modules.base import BaseModule
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.modules.util import as_task
from gajim.common.modules.util import event_node


class UserAvatar(BaseModule):
    _nbxmpp_extends = "UserAvatar"
    _nbxmpp_methods = [
        "request_avatar_metadata",
        "request_avatar_data",
        "set_avatar",
        "set_access_model",
    ]

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)
        self._register_pubsub_handler(self._avatar_metadata_received)

    @event_node(Namespace.AVATAR_METADATA)
    def _avatar_metadata_received(
        self, _con: types.NBXMPPClient, _stanza: Message, properties: MessageProperties
    ) -> None:
        assert properties.pubsub_event is not None
        if properties.pubsub_event.retracted:
            return

        metadata = properties.pubsub_event.data
        assert properties.jid is not None
        jid = properties.jid
        contact = self._con.get_module("Contacts").get_contact(jid)
        assert isinstance(
            contact, BareContact | GroupchatContact | GroupchatParticipant
        )

        if metadata is None or not metadata.infos:
            self._log.info("No avatar published: %s", jid)
            app.storage.archive.set_contact_value(
                self._account, contact.jid, "avatar_sha", None
            )
            contact.update_avatar(None)
            return

        assert isinstance(metadata, AvatarMetaData)
        assert metadata.default is not None
        sha = contact.avatar_sha
        if sha is not None:
            if sha in metadata.avatar_shas and app.app.avatar_storage.avatar_exists(
                sha
            ):
                self._log.info("Avatar already known: %s %s", jid, sha)
                return

        if app.app.avatar_storage.avatar_exists(metadata.default):
            self._log.info(
                "Avatar found in cache, update: %s %s", jid, metadata.default
            )
            app.storage.archive.set_contact_value(
                self._account, contact.jid, "avatar_sha", metadata.default
            )
            contact.update_avatar(metadata.default)
            return

        # There are following cases this code is reached:
        # - We lost the avatar cache
        # - We use an outdated avatar
        # - We currently have no avatar set
        #
        # Reset the sha, because we don’t know if the avatar data query will
        # succeed. This forces an update of the avatar if the query succeeds.
        app.storage.archive.set_contact_value(
            self._account, contact.jid, "avatar_sha", None
        )
        contact.update_avatar(None)
        self._request_avatar_data(contact, metadata.default)

    @as_task
    def _request_avatar_data(
        self, contact: types.ChatContactT, sha: str
    ) -> Generator[Any, Any]:

        self._log.info("Request: %s %s", contact.jid, sha)

        _task = yield

        avatar = yield self._nbxmpp("UserAvatar").request_avatar_data(
            sha, jid=contact.jid
        )

        if avatar is None:
            self._log.warning(
                "%s advertised %s but data node is empty", contact.jid, sha
            )
            return

        if is_error(avatar):
            self._log.warning(avatar)
            return

        assert isinstance(avatar, AvatarData)
        self._log.info("Received Avatar: %s %s", contact.jid, avatar.sha)
        app.app.avatar_storage.save_avatar(avatar.data)
        app.storage.archive.set_contact_value(
            self._account, contact.jid, "avatar_sha", avatar.sha
        )
        contact.update_avatar(avatar.sha)
